import csv
import os
import threading
import traceback

from flask import g

from sqlalchemy.orm import joinedload

import chellow.dloads
from chellow.models import Llfc, Session
from chellow.utils import hh_format
from chellow.views import chellow_redirect


def content(user):
    sess = f = writer = None
    try:
        sess = Session()
        running_name, finished_name = chellow.dloads.make_names("llfcs.csv", user)
        f = open(running_name, mode="w", newline="")
        writer = csv.writer(f, lineterminator="\n")
        writer.writerow(
            (
                "Chellow Id",
                "DNO Code",
                "Code",
                "Description",
                "Voltage Level",
                "Is Substation?",
                "Is Import?",
                "Valid From",
                "Valid To",
            )
        )

        for llfc in (
            sess.query(Llfc)
            .order_by(Llfc.id)
            .options(joinedload(Llfc.dno), joinedload(Llfc.voltage_level))
        ):
            writer.writerow(
                (
                    str(llfc.id),
                    llfc.dno.dno_code,
                    llfc.code,
                    llfc.description,
                    llfc.voltage_level.code,
                    llfc.is_substation,
                    llfc.is_import,
                    hh_format(llfc.valid_from),
                    hh_format(llfc.valid_to),
                )
            )
    except BaseException:
        writer.writerow([traceback.format_exc()])
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_get(sess):
    args = (g.user,)
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



import re
from itertools import zip_longest
from time import sleep


def match(response, status_code, *patterns):
    response_str = response.get_data(as_text=True)

    assert response.status_code == status_code, response_str

    for regex in patterns:
        assert re.search(
            regex, response_str, flags=re.MULTILINE + re.DOTALL
        ), response_str


def match_tables(table_1, table_2):
    for r1, r2 in zip_longest(table_1, table_2):
        for c1, c2 in zip_longest(r1, r2):
            if c1 != c2:
                raise Exception(f"Two cells don't match: {c1} and {c2}")

    assert table_1 == table_2


def match_repeat(client, path, match, seconds=5):
    for second in range(seconds):
        response = client.get(path)
        response_str = response.get_data(as_text=True)
        if match in response_str:
            return response
        elif second == seconds - 1:
            assert match in response_str, response_str
        else:
            sleep(1)



import atexit
import collections
import threading
import traceback
from base64 import b64decode
from importlib import import_module

import requests

from werkzeug.exceptions import BadRequest

from chellow.models import (
    Contract,
    Session,
)
from chellow.utils import utc_datetime_now


importer = None


def download(s, url):
    fl_json = s.get(url).json()
    return b64decode(fl_json["content"])


def run_import(sess, log, set_progress):
    log("Starting to import rates from the rate server")
    conf = Contract.get_non_core_by_name(sess, "configuration")
    props = conf.make_properties()
    repo_props = props.get("rate_server", {})
    repo_url = repo_props.get(
        "url", "https://api.github.com/repos/WessexWater/chellow-rates"
    )
    repo_branch = repo_props.get("branch", "main")
    s = requests.Session()
    s.verify = False
    repo_res = s.get(repo_url)
    try:
        repo_entry = repo_res.json()
    except requests.exceptions.JSONDecodeError as e:
        raise BadRequest(
            f"Couldn't parse as JSON the content from {repo_url} with error {e}: "
            f"{repo_res.text}"
        )
    if "message" in repo_entry:
        raise Exception(f"Message from the GitHub API: {repo_entry['message']}")

    log(
        f"Looking at {repo_entry['html_url']} and branch "
        f"{'default' if repo_branch is None else repo_branch}"
    )
    branch_entry = s.get(f"{repo_url}/branches/{repo_branch}").json()

    tree_entry = s.get(
        branch_entry["commit"]["commit"]["tree"]["url"],
        params={"recursive": "true"},
    ).json()

    if tree_entry["truncated"]:
        raise Exception("Tree from rate server is truncated.")

    paths_list = []
    for sub_entry in tree_entry["tree"]:
        path = sub_entry["path"].split("/")
        if path[-1] != "README.md":
            paths_list.append((path, sub_entry["url"]))

    paths = tuple(paths_list)

    for mod_name in (
        "chellow.e.bsuos",
        "chellow.e.dno_rate_parser",
        "chellow.e.laf_import",
        "chellow.e.mdd_importer",
        "chellow.e.tnuos",
        "chellow.gas.dn_rate_parser",
    ):
        mod = import_module(mod_name)
        mod.rate_server_import(sess, log, set_progress, s, paths)


class RateServer(threading.Thread):
    def __init__(self):
        super().__init__(name="Rate Server")
        self.lock = threading.RLock()
        self.messages = collections.deque(maxlen=500)
        self.progress = ""
        self.stopped = threading.Event()
        self.going = threading.Event()

    def stop(self):
        self.stopped.set()
        self.going.set()
        self.join()

    def go(self):
        self.going.set()

    def is_locked(self):
        if self.lock.acquire(False):
            self.lock.release()
            return False
        else:
            return True

    def log(self, message):
        self.messages.appendleft(
            f"{utc_datetime_now().strftime('%Y-%m-%d %H:%M:%S')} - {message}"
        )

    def set_progress(self, progress):
        self.progress = progress

    def run(self):
        while not self.stopped.isSet():
            if self.lock.acquire(False):
                sess = self.global_alert = None
                try:
                    sess = Session()
                    run_import(sess, self.log, self.set_progress)
                except BaseException as e:
                    msg = f"{e.description} " if isinstance(e, BadRequest) else ""
                    self.log(f"{msg}{traceback.format_exc()}")
                    self.global_alert = (
                        "There's a problem with a <a href='/rate_server'>Rate Server "
                        "import</a>."
                    )
                    sess.rollback()
                finally:
                    if sess is not None:
                        sess.close()
                    self.lock.release()
                    self.log("Finished importing rates.")

            self.going.wait(60 * 60 * 24)
            self.going.clear()


def get_importer():
    return importer


def startup():
    global importer
    importer = RateServer()
    importer.start()


@atexit.register
def shutdown():
    if importer is not None:
        importer.stop()



from decimal import Decimal, InvalidOperation

from openpyxl import load_workbook

from werkzeug.exceptions import BadRequest

from chellow.models import Session
from chellow.utils import ct_datetime, parse_mpan_core, to_utc


def get_ct_date(title_row, row, name):
    return get_value(title_row, row, name)


def get_start_date(title_row, row, name):
    return to_utc(get_ct_date(title_row, row, name))


def get_finish_date(title_row, row, name):
    d = get_ct_date(title_row, row, name)

    return to_utc(ct_datetime(d.year, d.month, d.day, 23, 30))


def get_value(title_row, row, name):
    idx = None
    name = name.strip().lower()

    for i, title_cell in enumerate(title_row):
        if str(title_cell.value).strip().lower() == name:
            idx = i
            break

    if idx is None:
        raise BadRequest(
            "The name '{name}' can't be found in the titles "
            "{title_row}.".format(name=name, title_row=title_row)
        )

    try:
        return row[idx].value

    except IndexError:
        raise BadRequest(
            f"For the row {row}, the index is {idx} which is beyond the end of the row."
        )


def get_str(title_row, row, name):
    return get_value(title_row, row, name).strip()


def get_dec(title_row, row, name):
    try:
        return Decimal(str(get_value(title_row, row, name)))
    except InvalidOperation:
        return None


def get_int(title_row, row, name):
    return int(get_value(title_row, row, name))


METER_RATE = Decimal("60.00")


class Parser:
    def __init__(self, f):
        self.book = load_workbook(f)
        self.sheet = self.book.worksheets[0]

        self.last_line = None
        self._line_number = None
        self._title_line = None

    @property
    def line_number(self):
        return None if self._line_number is None else self._line_number + 1

    def _set_last_line(self, i, line):
        self._line_number = i
        self.last_line = line
        if i == 0:
            self._title_line = line
        return line

    def make_raw_bills(self):
        row_index = None
        sess = None
        try:
            sess = Session()
            bills = []
            row_iter = self.sheet.iter_rows()
            title_row = next(row_iter)
            for row in row_iter:
                val = get_value(title_row, row, "mpan ref")
                if val is None or val == "":
                    break

                self._set_last_line(row_index, val)
                msn = str(get_value(title_row, row, "meter")).strip()
                mpan_core = parse_mpan_core(str(get_int(title_row, row, "mpan ref")))
                start_date = get_start_date(title_row, row, "start")
                issue_date = start_date
                finish_date = get_finish_date(title_row, row, "end")
                check = get_str(title_row, row, "check")
                if check != "Billed":
                    continue

                net = METER_RATE / 12
                vat = round(net * Decimal("0.2"), 2)

                breakdown = {
                    "raw_lines": [],
                    "cop": ["5"],
                    "settlement-status": ["non_settlement"],
                    "msn": [msn],
                    "meter-rate": [METER_RATE],
                    "meter-gbp": net,
                }

                bills.append(
                    {
                        "bill_type_code": "N",
                        "kwh": Decimal(0),
                        "vat": vat,
                        "net": net,
                        "gross": net + vat,
                        "reads": [],
                        "breakdown": breakdown,
                        "account": mpan_core,
                        "issue_date": issue_date,
                        "start_date": start_date,
                        "finish_date": finish_date,
                        "mpan_core": mpan_core,
                        "reference": "_".join(
                            (
                                start_date.strftime("%Y%m%d"),
                                finish_date.strftime("%Y%m%d"),
                                issue_date.strftime("%Y%m%d"),
                                mpan_core,
                            )
                        ),
                    }
                )
                sess.rollback()
        except BadRequest as e:
            raise BadRequest("Row number: " + str(row_index) + " " + e.description)
        finally:
            if sess is not None:
                sess.close()

        return bills



import atexit
import collections
import datetime
import threading
import traceback
from datetime import timedelta as Timedelta

import requests

from werkzeug.exceptions import BadRequest

import xlrd

from zish import loads

from chellow.models import Contract, RateScript, Session, get_non_core_contract_id
from chellow.utils import HH, hh_format, to_ct, to_utc, utc_datetime_now


ELEXON_PORTAL_SCRIPTING_KEY_KEY = "elexonportal_scripting_key"


def key_format(dt):
    return dt.strftime("%d %H:%M")


def hh(data_source):
    for h in data_source.hh_data:
        try:
            sbp, ssp = data_source.caches["system_price"][h["start-date"]]
        except KeyError:
            try:
                system_price_cache = data_source.caches["system_price"]
            except KeyError:
                system_price_cache = data_source.caches["system_price"] = {}

            db_id = get_non_core_contract_id("system_price")
            h_start = h["start-date"]
            rates = data_source.hh_rate(db_id, h_start)["gbp_per_nbp_mwh"]

            try:
                try:
                    rdict = rates[key_format(h_start)]
                except KeyError:
                    rdict = rates[key_format(h_start - Timedelta(days=3))]
                sbp = float(rdict["sbp"] / 1000)
                ssp = float(rdict["ssp"] / 1000)
                system_price_cache[h_start] = (sbp, ssp)
            except KeyError:
                raise BadRequest(
                    f"For the System Price rate script at {hh_format(h_start)} "
                    f"the rate cannot be found."
                )
            except TypeError:
                raise BadRequest(
                    f"For the System Price rate script at {hh_format(h_start)} "
                    f"the rate 'rates_gbp_per_mwh' has the problem: "
                    f"{traceback.format_exc()}"
                )

        h["sbp"] = sbp
        h["sbp-gbp"] = h["nbp-kwh"] * sbp

        h["ssp"] = ssp
        h["ssp-gbp"] = h["nbp-kwh"] * ssp


system_price_importer = None


class SystemPriceImporter(threading.Thread):
    def __init__(self):
        super(SystemPriceImporter, self).__init__(name="System Price Importer")
        self.lock = threading.RLock()
        self.messages = collections.deque(maxlen=1000)
        self.stopped = threading.Event()
        self.going = threading.Event()
        self.PROXY_HOST_KEY = "proxy.host"
        self.PROXY_PORT_KEY = "proxy.port"

    def stop(self):
        self.stopped.set()
        self.going.set()
        self.join()

    def go(self):
        self.going.set()

    def is_locked(self):
        if self.lock.acquire(False):
            self.lock.release()
            return False
        else:
            return True

    def log(self, message):
        self.messages.appendleft(
            utc_datetime_now().strftime("%Y-%m-%d %H:%M:%S") + " - " + message
        )

    def run(self):
        while not self.stopped.isSet():
            if self.lock.acquire(False):
                sess = None
                try:
                    sess = Session()
                    _process(self.log, sess)
                except BaseException:
                    self.log(f"Outer problem {traceback.format_exc()}")
                    sess.rollback()
                finally:
                    self.lock.release()
                    self.log("Finished checking System Price rates.")
                    if sess is not None:
                        sess.close()

            self.going.wait(24 * 60 * 60)
            self.going.clear()


def _process(log_f, sess):
    contract = Contract.get_non_core_by_name(sess, "system_price")
    contract_props = contract.make_properties()

    if not contract_props.get("enabled", False):
        log_f(
            "The automatic importer is disabled. To enable it, edit "
            "the contract properties to set 'enabled' to True."
        )
        return

    log_f("Starting to check System Prices.")

    for rscript in (
        sess.query(RateScript)
        .filter(RateScript.contract == contract)
        .order_by(RateScript.start_date.desc())
    ):
        ns = loads(rscript.script)
        rates = ns["gbp_per_nbp_mwh"]
        if len(rates) == 0:
            fill_start = rscript.start_date
            break
        elif rates[key_format(rscript.finish_date)]["run"] == "DF":
            fill_start = rscript.finish_date + HH
            break

    config = Contract.get_non_core_by_name(sess, "configuration")
    config_props = config.make_properties()

    scripting_key = config_props.get(ELEXON_PORTAL_SCRIPTING_KEY_KEY)
    if scripting_key is None:
        raise BadRequest(
            f"The property {ELEXON_PORTAL_SCRIPTING_KEY_KEY} cannot be found in "
            f"the configuration properties."
        )
    url = (
        f"{contract_props['url']}file/download/BESTVIEWPRICES_FILE?key={scripting_key}"
    )

    log_f(f"Downloading from {url} and extracting data from {hh_format(fill_start)}")

    sess.rollback()  # Avoid long-running transactions
    res = requests.get(url)
    log_f(f"Received {res.status_code} {res.reason}")
    data = res.content
    book = xlrd.open_workbook(file_contents=data)
    sbp_sheet = book.sheet_by_index(1)
    ssp_sheet = book.sheet_by_index(2)

    sp_months = []
    sp_month = None
    for row_index in range(1, sbp_sheet.nrows):
        sbp_row = sbp_sheet.row(row_index)
        ssp_row = ssp_sheet.row(row_index)
        raw_date = datetime.datetime(
            *xlrd.xldate_as_tuple(sbp_row[0].value, book.datemode)
        )
        hh_date_ct = to_ct(raw_date)
        hh_date = to_utc(hh_date_ct)
        run_code = sbp_row[1].value
        for col_idx in range(2, 52):
            if hh_date >= fill_start:
                sbp_val = sbp_row[col_idx].value
                if sbp_val != "":
                    if hh_date.day == 1 and hh_date.hour == 0 and hh_date.minute == 0:
                        sp_month = {}
                        sp_months.append(sp_month)
                    ssp_val = ssp_row[col_idx].value
                    sp_month[hh_date] = {
                        "run": run_code,
                        "sbp": sbp_val,
                        "ssp": ssp_val,
                    }
            hh_date += HH
    log_f("Successfully extracted data.")
    last_date = sorted(sp_months[-1].keys())[-1]
    if last_date.month == (last_date + HH).month:
        del sp_months[-1]
    if "limit" in contract_props:
        sp_months = sp_months[0:1]
    for sp_month in sp_months:
        sorted_keys = sorted(sp_month.keys())
        month_start = sorted_keys[0]
        month_finish = sorted_keys[-1]
        rs = (
            sess.query(RateScript)
            .filter(
                RateScript.contract == contract,
                RateScript.start_date == month_start,
            )
            .first()
        )
        if rs is None:
            log_f(f"Adding a new rate script starting at {hh_format(month_start)}.")

            latest_rs = (
                sess.query(RateScript)
                .filter(RateScript.contract == contract)
                .order_by(RateScript.start_date.desc())
                .first()
            )

            contract.update_rate_script(
                sess,
                latest_rs,
                latest_rs.start_date,
                month_finish,
                loads(latest_rs.script),
            )
            rs = contract.insert_rate_script(sess, month_start, {})
            sess.flush()
        script = {
            "gbp_per_nbp_mwh": dict((key_format(k), v) for k, v in sp_month.items())
        }
        log_f(f"Updating rate script starting at {hh_format(month_start)}.")
        contract.update_rate_script(sess, rs, rs.start_date, rs.finish_date, script)
        sess.commit()


def get_importer():
    return system_price_importer


def startup():
    global system_price_importer
    system_price_importer = SystemPriceImporter()
    system_price_importer.start()


@atexit.register
def shutdown():
    if system_price_importer is not None:
        system_price_importer.stop()



import os
import sys
import threading
import traceback

from flask import g, request

import odio

from sqlalchemy import or_, true
from sqlalchemy.orm import joinedload
from sqlalchemy.sql.expression import null

from werkzeug.exceptions import BadRequest

import chellow.dloads
import chellow.e.computer
from chellow.e.computer import contract_func
from chellow.gas.engine import GDataSource
from chellow.models import (
    GBill,
    GContract,
    GEra,
    GSupply,
    Session,
    Site,
    SiteGEra,
)
from chellow.utils import (
    c_months_u,
    ct_datetime_now,
    hh_format,
    hh_max,
    hh_min,
    make_val,
    req_bool,
    req_int,
)
from chellow.views import chellow_redirect


CATEGORY_ORDER = {None: 0, "unmetered": 1, "nhh": 2, "amr": 3, "hh": 4}
meter_order = {"hh": 0, "amr": 1, "nhh": 2, "unmetered": 3}


def write_spreadsheet(fl, compressed, site_rows, era_rows):
    fl.seek(0)
    fl.truncate()
    with odio.create_spreadsheet(fl, "1.2", compressed=compressed) as f:
        f.append_table("Site Level", site_rows)
        f.append_table("Era Level", era_rows)


def content(
    site_id, g_supply_id, user, compression, finish_year, finish_month, months, now=None
):
    if now is None:
        now = ct_datetime_now()
    report_context = {}
    sess = None
    month_list = list(
        c_months_u(finish_year=finish_year, finish_month=finish_month, months=months)
    )
    start_date, finish_date = month_list[0][0], month_list[-1][-1]

    try:
        sess = Session()
        base_name = [
            "g_monthly_duration",
            hh_format(start_date).replace(" ", "_").replace(":", "").replace("-", ""),
            "for",
            str(months),
            "months",
        ]

        forecast_from = chellow.e.computer.forecast_date()

        sites = (
            sess.query(Site)
            .join(SiteGEra)
            .join(GEra)
            .filter(SiteGEra.is_physical == true())
            .distinct()
            .order_by(Site.code)
        )
        if site_id is not None:
            site = Site.get_by_id(sess, site_id)
            sites = sites.filter(Site.id == site.id)
            base_name.append("site")
            base_name.append(site.code)
        if g_supply_id is not None:
            g_supply = GSupply.get_by_id(sess, g_supply_id)
            base_name.append("g_supply")
            base_name.append(str(g_supply.id))
            sites = sites.filter(GEra.g_supply == g_supply)

        running_name, finished_name = chellow.dloads.make_names(
            "_".join(base_name) + ".ods", user
        )

        rf = open(running_name, "wb")
        site_rows = []
        g_era_rows = []

        era_header_titles = [
            "creation_date",
            "mprn",
            "supply_name",
            "exit_zone",
            "msn",
            "unit",
            "contract",
            "site_id",
            "site_name",
            "associated_site_ids",
            "month",
        ]
        site_header_titles = [
            "creation_date",
            "site_id",
            "site_name",
            "associated_site_ids",
            "month",
        ]
        summary_titles = ["kwh", "gbp", "billed_kwh", "billed_gbp"]

        vb_titles = []
        conts = (
            sess.query(GContract)
            .join(GEra)
            .join(GSupply)
            .filter(
                GEra.start_date <= finish_date,
                or_(GEra.finish_date == null(), GEra.finish_date >= start_date),
            )
            .distinct()
            .order_by(GContract.id)
        )
        if g_supply_id is not None:
            conts = conts.filter(GEra.g_supply_id == g_supply_id)
        for cont in conts:
            title_func = chellow.e.computer.contract_func(
                report_context, cont, "virtual_bill_titles"
            )
            if title_func is None:
                raise Exception(
                    f"For the contract {cont.name} there doesn't seem to be "
                    f"a 'virtual_bill_titles' function."
                )
            for title in title_func():
                if title not in vb_titles:
                    vb_titles.append(title)

        g_era_rows.append(era_header_titles + summary_titles + vb_titles)
        site_rows.append(site_header_titles + summary_titles)

        for month_start, month_finish in month_list:
            for site in sites.filter(
                GEra.start_date <= month_finish,
                or_(GEra.finish_date == null(), GEra.finish_date >= month_start),
            ):
                site_kwh = site_gbp = site_billed_kwh = site_billed_gbp = 0

                for g_era in (
                    sess.query(GEra)
                    .join(SiteGEra)
                    .filter(
                        SiteGEra.site == site,
                        SiteGEra.is_physical == true(),
                        GEra.start_date <= month_finish,
                        or_(
                            GEra.finish_date == null(), GEra.finish_date >= month_start
                        ),
                    )
                    .options(
                        joinedload(GEra.g_contract),
                        joinedload(GEra.g_supply),
                        joinedload(GEra.g_supply).joinedload(GSupply.g_exit_zone),
                    )
                    .order_by(GEra.id)
                ):
                    g_supply = g_era.g_supply

                    if g_supply_id is not None and g_supply.id != g_supply_id:
                        continue

                    ss_start = hh_max(g_era.start_date, month_start)
                    ss_finish = hh_min(g_era.finish_date, month_finish)

                    ss = GDataSource(
                        sess,
                        ss_start,
                        ss_finish,
                        forecast_from,
                        g_era,
                        report_context,
                        None,
                    )

                    contract = g_era.g_contract
                    vb_function = contract_func(
                        report_context, contract, "virtual_bill"
                    )
                    if vb_function is None:
                        raise BadRequest(
                            "The contract "
                            + contract.name
                            + " doesn't have the virtual_bill() function."
                        )
                    vb_function(ss)
                    bill = ss.bill

                    try:
                        gbp = bill["net_gbp"]
                    except KeyError:
                        gbp = 0
                        bill["problem"] += (
                            "For the supply "
                            + ss.mprn
                            + " the virtual bill "
                            + str(bill)
                            + " from the contract "
                            + contract.name
                            + " does not contain the net_gbp key."
                        )
                    try:
                        kwh = bill["kwh"]
                    except KeyError:
                        kwh = 0
                        bill["problem"] += (
                            "For the supply "
                            + ss.mprn
                            + " the virtual bill "
                            + str(bill)
                            + " from the contract "
                            + contract.name
                            + " does not contain the 'kwh' key."
                        )

                    billed_kwh = billed_gbp = 0

                    g_era_associates = {
                        s.site.code for s in g_era.site_g_eras if not s.is_physical
                    }

                    for g_bill in sess.query(GBill).filter(
                        GBill.g_supply == g_supply,
                        GBill.start_date <= ss_finish,
                        GBill.finish_date >= ss_start,
                    ):
                        bill_start = g_bill.start_date
                        bill_finish = g_bill.finish_date
                        bill_duration = (bill_finish - bill_start).total_seconds() + (
                            30 * 60
                        )
                        overlap_duration = (
                            min(bill_finish, ss_finish) - max(bill_start, ss_start)
                        ).total_seconds() + (30 * 60)
                        overlap_proportion = overlap_duration / bill_duration
                        billed_kwh += overlap_proportion * float(g_bill.kwh)
                        billed_gbp += overlap_proportion * float(g_bill.net)

                    associated_site_ids = ",".join(sorted(g_era_associates))
                    g_era_rows.append(
                        [
                            make_val(v)
                            for v in [
                                now,
                                g_supply.mprn,
                                g_supply.name,
                                g_supply.g_exit_zone.code,
                                g_era.msn,
                                g_era.g_unit.code,
                                contract.name,
                                site.code,
                                site.name,
                                associated_site_ids,
                                month_finish,
                                kwh,
                                gbp,
                                billed_kwh,
                                billed_gbp,
                            ]
                        ]
                        + [make_val(bill.get(t)) for t in vb_titles]
                    )

                    site_kwh += kwh
                    site_gbp += gbp
                    site_billed_kwh += billed_kwh
                    site_billed_gbp += billed_gbp

                linked_sites = ", ".join(
                    s.code
                    for s in site.find_linked_sites(sess, month_start, month_finish)
                )

                site_rows.append(
                    [
                        make_val(v)
                        for v in [
                            now,
                            site.code,
                            site.name,
                            linked_sites,
                            month_finish,
                            site_kwh,
                            site_gbp,
                            site_billed_kwh,
                            site_billed_gbp,
                        ]
                    ]
                )
                sess.rollback()
            write_spreadsheet(rf, compression, site_rows, g_era_rows)

    except BadRequest as e:
        site_rows.append(["Problem " + e.description])
        write_spreadsheet(rf, compression, site_rows, g_era_rows)
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg + "\n")
        site_rows.append(["Problem " + msg])
        write_spreadsheet(rf, compression, site_rows, g_era_rows)
    finally:
        if sess is not None:
            sess.close()
        try:
            rf.close()
            os.rename(running_name, finished_name)
        except BaseException:
            msg = traceback.format_exc()
            r_name, f_name = chellow.dloads.make_names("error.txt", user)
            ef = open(r_name, "w")
            ef.write(msg + "\n")
            ef.close()


def do_get(sess):
    finish_year = req_int("finish_year")
    finish_month = req_int("finish_month")
    months = req_int("months")

    site_id = req_int("site_id") if "site_id" in request.values else None

    if "g_supply_id" in request.values:
        g_supply_id = req_int("g_supply_id")
    else:
        g_supply_id = None

    if "compression" in request.values:
        compression = req_bool("compression")
    else:
        compression = True

    user = g.user
    args = (site_id, g_supply_id, user, compression, finish_year, finish_month, months)

    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



from collections import defaultdict
from decimal import Decimal
from io import BytesIO


from chellow.e.bill_parsers.haven_edi_tprs import (
    Parser,
    _process_CCD1,
    _process_CCD3,
    _process_MAN,
    _process_MHD,
    _process_MTR,
)
from chellow.utils import utc_datetime


def test_process_MTR_UTLHDR(mocker):
    elements = {}
    headers = {"message_type": "UTLHDR", "breakdown": {}}
    bill = _process_MTR(elements, headers)
    assert bill is None


def test_process_MTR_UTLBIL(mocker):
    gbp = "10.31"
    cons = "113"
    elements = {}
    headers = {
        "is_ebatch": False,
        "message_type": "UTLBIL",
        "breakdown": {
            "00206-gbp": Decimal("10.31"),
            "00206-kwh": Decimal("113"),
            "00206-rate": {Decimal("0.0001")},
        },
        "mpan_core": ["0850"],
        "kwh": 8,
        "reference": "a",
        "issue_date": "d",
        "account": "acc",
        "start_date": "d",
        "finish_date": "d",
        "net": 0,
        "vat": 0,
        "gross": 0,
        "reads": [
            {
                "msn": "hgkh",
                "mpan": "      ",
                "coefficient": Decimal("0.00001"),
                "units": "kWh",
                "tpr_code": "00040",
                "prev_date": utc_datetime(2020, 3, 31, 22, 30),
                "prev_value": Decimal("1"),
                "prev_type_code": "N",
                "pres_date": utc_datetime(2020, 3, 1, 23, 30),
                "pres_value": Decimal("0"),
                "pres_type_code": "N",
            },
        ],
        "bill_type_code": "N",
    }
    expected_headers = {
        "is_ebatch": False,
        "message_type": "UTLBIL",
        "kwh": 8,
        "reference": "a",
        "mpan_core": ["0850"],
        "issue_date": "d",
        "account": "acc",
        "start_date": "d",
        "finish_date": "d",
        "net": 0,
        "vat": 0,
        "gross": 0,
        "breakdown": {
            "00206-gbp": Decimal(gbp),
            "00206-rate": {Decimal("0.0001")},
            "00206-kwh": Decimal(cons),
        },
        "reads": [
            {
                "msn": "hgkh",
                "mpan": "      ",
                "coefficient": Decimal("0.00001"),
                "units": "kWh",
                "tpr_code": "00040",
                "prev_date": utc_datetime(2020, 3, 31, 22, 30),
                "prev_value": Decimal("1"),
                "prev_type_code": "N",
                "pres_date": utc_datetime(2020, 3, 1, 23, 30),
                "pres_value": Decimal("0"),
                "pres_type_code": "N",
            }
        ],
        "bill_type_code": "N",
    }
    _process_MTR(elements, headers)
    assert headers == expected_headers


def test_process_MTR_UTLBIL_multiple_charges_one_tpr(mocker):
    elements = {}
    headers = {
        "is_ebatch": False,
        "message_type": "UTLBIL",
        "breakdown": {},
        "mpan_core": ["0850"],
        "kwh": 8,
        "reference": "a",
        "issue_date": "d",
        "account": "acc",
        "start_date": "d",
        "finish_date": "d",
        "net": 0,
        "vat": 0,
        "gross": 0,
        "reads": [
            {
                "msn": "hgkh",
                "mpan": "      ",
                "coefficient": Decimal("0.00001"),
                "units": "kWh",
                "tpr_code": "00040",
                "prev_date": utc_datetime(2020, 3, 31, 22, 30),
                "prev_value": Decimal("1"),
                "prev_type_code": "N",
                "pres_date": utc_datetime(2020, 3, 1, 23, 30),
                "pres_value": Decimal("0"),
                "pres_type_code": "N",
            },
        ],
        "bill_type_code": "N",
    }
    expected_headers = {
        "is_ebatch": False,
        "message_type": "UTLBIL",
        "kwh": 8,
        "reference": "a",
        "mpan_core": ["0850"],
        "issue_date": "d",
        "account": "acc",
        "start_date": "d",
        "finish_date": "d",
        "net": 0,
        "vat": 0,
        "gross": 0,
        "breakdown": {},
        "reads": [
            {
                "msn": "hgkh",
                "mpan": "      ",
                "coefficient": Decimal("0.00001"),
                "units": "kWh",
                "tpr_code": "00040",
                "prev_date": utc_datetime(2020, 3, 31, 22, 30),
                "prev_value": Decimal("1"),
                "prev_type_code": "N",
                "pres_date": utc_datetime(2020, 3, 1, 23, 30),
                "pres_value": Decimal("0"),
                "pres_type_code": "N",
            }
        ],
        "bill_type_code": "N",
    }
    _process_MTR(elements, headers)
    assert headers == expected_headers


def test_process_MTR_UTLBIL_unmetered(mocker):
    elements = {}
    headers = {
        "is_ebatch": False,
        "message_type": "UTLBIL",
        "breakdown": {},
        "mpan_core": ["0850"],
        "kwh": 8,
        "reference": "a",
        "issue_date": "d",
        "account": "acc",
        "start_date": "d",
        "finish_date": "d",
        "net": 0,
        "vat": 0,
        "gross": 0,
        "reads": [],
        "bill_type_code": "N",
    }
    expected_headers = {
        "is_ebatch": False,
        "message_type": "UTLBIL",
        "kwh": 8,
        "reference": "a",
        "mpan_core": ["0850"],
        "issue_date": "d",
        "account": "acc",
        "start_date": "d",
        "finish_date": "d",
        "net": 0,
        "vat": 0,
        "gross": 0,
        "breakdown": {},
        "reads": [],
        "bill_type_code": "N",
    }
    _process_MTR(elements, headers)
    assert headers == expected_headers


def test_process_MAN(mocker):
    elements = {
        "MADN": ["20", "0000000000", "6", "00", "001", "002"],
    }

    headers = {}
    _process_MAN(elements, headers)
    expected_headers = {"mpan_core": "20 0000 0000 006"}
    assert headers == expected_headers


def test_process_MHD(mocker):
    message_type = "UTLBIL"
    elements = {"TYPE": [message_type]}

    sess = mocker.Mock()
    headers = {"sess": sess}
    _process_MHD(elements, headers)
    expected_headers = {
        "message_type": message_type,
        "reads": [],
        "errors": [],
        "bill_elements": [],
        "breakdown": defaultdict(int, {"raw-lines": []}),
        "sess": sess,
        "kwh": Decimal("0"),
        "account": None,
        "start_date": None,
        "finish_date": None,
        "gross": Decimal("0.00"),
        "vat": Decimal("0.00"),
        "net": Decimal("0.00"),
        "reference": None,
        "mpan_core": None,
        "is_ebatch": False,
        "issue_date": None,
        "bill_type_code": None,
    }
    assert headers == expected_headers
    assert isinstance(headers["breakdown"], type(expected_headers["breakdown"]))


def test_process_CCD3(mocker):
    elements = {
        "CCDE": ["3", "", "NRG"],
        "TCOD": ["NIGHT", "Night"],
        "TMOD": ["453043"],
        "CONS": [[]],
        "BPRI": ["10"],
    }

    headers = {"kwh": Decimal("0"), "breakdown": {}}

    _process_CCD3(elements, headers)

    expected_headers = {
        "kwh": Decimal("0"),
        "breakdown": {},
    }
    assert headers == expected_headers


def test_process_CCD3_ebrs(mocker):
    elements = {
        "CCDE": ["3", "", "NRG"],
        "TMOD": ["823408"],
        "CPPU": ["028381"],
        "CTOT": ["47119"],
        "NUCT": ["4127137"],
    }

    headers = {"kwh": Decimal("0"), "breakdown": defaultdict(int, {})}

    _process_CCD3(elements, headers)

    expected_headers = {
        "kwh": Decimal("0"),
        "breakdown": {
            "ebrs-gbp": Decimal("471.19"),
            "ebrs-kwh": Decimal("4127.137"),
            "ebrs-rate": {Decimal("0.28381")},
        },
    }
    assert headers == expected_headers


def test_process_CCD3_ebrs_kwh(mocker):
    elements = {
        "CCDE": ["3", "", "NRG"],
        "TMOD": ["823408"],
        "NUCT": ["4127137"],
    }

    headers = {"kwh": Decimal("0"), "breakdown": defaultdict(int, {})}

    _process_CCD3(elements, headers)

    expected_headers = {
        "kwh": Decimal("0"),
        "breakdown": {
            "ebrs-kwh": Decimal("4127.137"),
        },
    }
    assert headers == expected_headers


def test_process_CCD_1(mocker):
    msn = "hgkh"

    elements = {
        "CCDE": ["1", "", "NRG"],
        "TCOD": ["NIGHT", "Night"],
        "TMOD": ["453043"],
        "MTNR": [msn],
        "CONS": [[]],
        "BPRI": ["10"],
        "PRDT": ["200301"],
        "PVDT": ["200331"],
        "MLOC": ["2275834732592"],
        "PRRD": ["0", "00", "1", "00"],
        "ADJF": ["", "1"],
    }

    headers = {
        "reads": [],
    }

    _process_CCD1(elements, headers)

    expected_headers = {
        "mpan_core": "22 7583 4732 592",
        "reads": [
            {
                "msn": msn,
                "mpan": "   22 7583 4732 592",
                "coefficient": Decimal("0.00001"),
                "units": "kWh",
                "tpr_code": "453043",
                "prev_date": utc_datetime(2020, 3, 31, 22, 30),
                "prev_value": Decimal("1"),
                "prev_type_code": "N",
                "pres_date": utc_datetime(2020, 3, 1, 23, 30),
                "pres_value": Decimal("0"),
                "pres_type_code": "N",
            }
        ],
    }
    assert headers == expected_headers


def test_process_CCD1_ebrs(mocker):
    elements = {
        "PRDT": ["200301"],
        "PVDT": ["200331"],
        "TCOD": ["EBRSD", "Energy Bill Relief Scheme Discount"],
        "TMOD": ["453043"],
        "MTNR": ["hgl"],
        "MLOC": ["2275834732592"],
        "PRRD": ["0", "00", "1", "00"],
        "ADJF": ["", "1"],
    }

    headers = {}

    _process_CCD1(elements, headers)

    expected_headers = {}
    assert headers == expected_headers


def test_Parser(mocker, sess):
    file_lines = [
        "MHD=2+UTLBIL:3'",
        "MTR=6'",
    ]
    f = BytesIO(b"\n".join(n.encode("utf8") for n in file_lines))
    parser = Parser(f)
    parser.make_raw_bills()



from decimal import Decimal

import chellow.general_import
from chellow.models import (
    Comm,
    Contract,
    Cop,
    EnergisationStatus,
    GspGroup,
    MarketRole,
    MeterPaymentType,
    MeterType,
    Mtc,
    MtcLlfc,
    MtcParticipant,
    Participant,
    Pc,
    Site,
    Source,
    VoltageLevel,
    insert_comms,
    insert_cops,
    insert_energisation_statuses,
    insert_sources,
    insert_voltage_levels,
)
from chellow.utils import ct_datetime, hh_format, to_utc, utc_datetime


def test_general_import_g_batch(mocker):
    sess = mocker.Mock()
    action = "insert"
    vals = ["CH4U", "batch 8883", "Apr 2019"]
    args = []
    chellow.general_import.general_import_g_batch(sess, action, vals, args)


def test_general_import_g_bill(mocker):
    c = mocker.patch("chellow.general_import.GContract", autospec=True)
    c.get_supplier_by_name.return_value = chellow.general_import.GContract(
        False, "CH4U", "{}", "{}"
    )

    mocker.patch("chellow.models.GBatch", autospec=True)
    batch = chellow.models.GBatch(1, 2, 3, 4)
    batch.g_contract = chellow.general_import.GContract(False, "CH4U", "{}", "{}")
    c.get_g_batch_by_reference.return_value = batch

    sess = mocker.Mock()
    action = "insert"
    vals = [
        "CH4U",
        "batch 8883",
        "759288812",
        "2019-09-08 00:00",
        "2019-10-01 00:00",
        "2019-10-31 23:30",
        "0.00",
        "0.00",
        "0.00",
        "77hwgtlll",
        "7876hrwlju",
        "N",
        "{}",
        "0",
    ]
    args = []
    chellow.general_import.general_import_g_bill(sess, action, vals, args)


def test_general_import_g_bill_reads(mocker):
    c = mocker.patch("chellow.general_import.GContract", autospec=True)
    contract = chellow.general_import.GContract(False, "CH4U", "{}", "{}")
    c.get_supplier_by_name.return_value = contract

    mocker.patch("chellow.models.GBatch", autospec=True)
    batch = chellow.models.GBatch("CH4U", "{}", "{}", 4)
    contract.get_g_batch_by_reference.return_value = batch

    mocker.patch("chellow.models.GBill", autospec=True)
    bill = chellow.models.GBill(
        mocker.Mock(),
        mocker.Mock(),
        mocker.Mock(),
        mocker.Mock(),
        mocker.Mock(),
        mocker.Mock(),
        mocker.Mock(),
        mocker.Mock(),
        mocker.Mock(),
        mocker.Mock(),
        mocker.Mock(),
        mocker.Mock(),
        mocker.Mock(),
        mocker.Mock(),
    )
    batch.insert_g_bill.return_value = bill

    sess = mocker.Mock()
    msn = "88hgkdshjf"
    g_unit_code = "M3"
    g_unit_class = mocker.patch("chellow.general_import.GUnit", autospec=True)
    g_unit = chellow.general_import.GUnit(g_unit_code, "", 1)
    g_unit_class.get_by_code.return_value = g_unit
    correction_factor = Decimal("1")
    calorific_value = Decimal("39")
    prev_value = Decimal("988")
    prev_date = utc_datetime(2019, 10, 1)
    prev_type_code = "E"
    pres_value = Decimal("1200")
    pres_date = utc_datetime(2019, 10, 31, 23, 30)
    pres_type_code = "A"

    g_read_type_class = mocker.patch("chellow.general_import.GReadType", autospec=True)
    prev_type = chellow.general_import.GReadType(prev_type_code, "")
    pres_type = chellow.general_import.GReadType(pres_type_code, "")
    g_read_type_class.get_by_code.side_effect = [prev_type, pres_type]

    action = "insert"
    vals = [
        "CH4U",
        "batch 8883",
        "759288812",
        "2019-09-08 01:00",
        "2019-10-01 01:00",
        "2019-10-31 23:30",
        "0.00",
        "0.00",
        "0.00",
        "77hwgtlll",
        "7876hrwlju",
        "N",
        "{}",
        "0",
        msn,
        g_unit_code,
        str(correction_factor),
        str(calorific_value),
        hh_format(prev_date),
        str(prev_value),
        prev_type_code,
        hh_format(pres_date),
        str(pres_value),
        pres_type_code,
    ]
    args = []
    chellow.general_import.general_import_g_bill(sess, action, vals, args)
    bill.insert_g_read.assert_called_with(
        sess,
        msn,
        g_unit,
        correction_factor,
        calorific_value,
        prev_value,
        prev_date,
        prev_type,
        pres_value,
        pres_date,
        pres_type,
    )


def test_parse_breakdown():
    breakdown_str = '{"date": 2009-05-12T03:00:00Z}'
    expected = {"date": utc_datetime(2009, 5, 12, 3)}
    actual = chellow.general_import._parse_breakdown(breakdown_str)
    assert actual == expected


def test_general_import_era_insert(sess):
    valid_from = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", valid_from, None, None)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess, "bank_holidays", "", {}, valid_from, None, bank_holiday_rate_script
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(
        sess, market_role_M, "Fusion Mop Ltd", valid_from, None, None
    )
    participant.insert_party(sess, market_role_X, "Fusion Ltc", valid_from, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", valid_from, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, valid_from, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, valid_from, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", utc_datetime(2000, 1, 1), None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    exp_supplier_contract = Contract.insert_supplier(
        sess, "Fusion Supplier 2000", participant, "", {}, valid_from, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", valid_from, None, "22")
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", utc_datetime(2000, 1, 1), None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", valid_from, None)
    mtc = Mtc.insert(sess, "845", False, True, valid_from, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        valid_from,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "521", "Export (HV)", voltage_level, False, False, valid_from, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, valid_from, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "net")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        "773",
        dc_contract,
        "ghyy3",
        "hgjeyhuw",
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        {},
        None,
        None,
        None,
        None,
        None,
        "22 7867 6232 781",
        "521",
        exp_supplier_contract,
        "7748",
        361,
    )

    sess.commit()

    action = "insert"
    vals = [
        "22 7867 6232 781",
        "2020-10-01 00:00",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "Fusion Supplier 2000",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
    ]
    args = []
    chellow.general_import.general_import_era(sess, action, vals, args)


def test_general_import_era_update(mocker):
    MockSupply = mocker.patch("chellow.general_import.Supply", autospec=True)

    mock_supply = mocker.Mock()

    MockSupply.get_by_mpan_core.return_value = mock_supply

    mock_era = mocker.Mock()
    mock_era.properties = "{}"

    mock_supply.find_era_at.return_value = mock_era

    sess = mocker.Mock()
    action = "update"
    vals = [
        "CH4U",
        "2019-09-08 00:00",
        "{no change}",
        "{no change}",
        "A Mop Contract",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
    ]
    args = []
    chellow.general_import.general_import_era(sess, action, vals, args)


def test_general_import_llfc_insert(sess):
    participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(
        sess, market_role_R, "WPD", utc_datetime(2000, 1, 1), None, "27"
    )
    insert_voltage_levels(sess)
    sess.commit()

    action = "insert"
    vals = [
        "27",
        "A1A",
        "LV:LV A Agg Band 0",
        "LV",
        "False",
        "True",
        "2020-10-21 00:00",
        "",
    ]
    args = []
    chellow.general_import.general_import_llfc(sess, action, vals, args)


def test_general_import_llfc_update(sess):
    participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    dno = participant.insert_party(
        sess, market_role_R, "WPD", utc_datetime(2000, 1, 1), None, "10"
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    dno.insert_llfc(
        sess,
        "328",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        to_utc(ct_datetime(2020, 4, 1)),
        None,
    )
    sess.commit()

    action = "update"
    vals = [
        "10",
        "328",
        "2020-04-01 00:00",
        "Reserved EHV 33kV - Import",
        "HV",
        "False",
        "True",
        "",
    ]
    args = []
    chellow.general_import.general_import_llfc(sess, action, vals, args)


def test_general_import_llfc_update_is_import_no_change(sess):
    participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    dno = participant.insert_party(
        sess, market_role_R, "WPD", utc_datetime(2000, 1, 1), None, "10"
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    dno.insert_llfc(
        sess,
        "328",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        to_utc(ct_datetime(2020, 4, 1)),
        None,
    )
    sess.commit()

    action = "update"
    vals = [
        "10",
        "328",
        "2020-04-01 00:00",
        "Reserved EHV 33kV - Import",
        "HV",
        "False",
        "{no change}",
        "",
    ]
    args = []
    chellow.general_import.general_import_llfc(sess, action, vals, args)


def test_general_import_llfc_update_valid_to_no_change(sess):
    participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    dno = participant.insert_party(
        sess, market_role_R, "WPD", utc_datetime(2000, 1, 1), None, "10"
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    dno.insert_llfc(
        sess,
        "328",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        to_utc(ct_datetime(2020, 4, 1)),
        None,
    )
    sess.commit()

    action = "update"
    vals = [
        "10",
        "328",
        "2020-04-01 00:00",
        "Reserved EHV 33kV - Import",
        "HV",
        "False",
        "{no change}",
        "{no change}",
    ]
    args = []
    chellow.general_import.general_import_llfc(sess, action, vals, args)



import csv
from datetime import datetime as Datetime, timedelta as Timedelta
from io import StringIO

from sqlalchemy import null, or_, select
from sqlalchemy.orm import joinedload

from werkzeug.exceptions import BadRequest

from chellow.models import (
    ClockInterval,
    Contract,
    GspGroup,
    Llfc,
    MarketRole,
    MeasurementRequirement,
    MeterPaymentType,
    MeterType,
    Mtc,
    MtcLlfc,
    MtcLlfcSsc,
    MtcLlfcSscPc,
    MtcParticipant,
    MtcSsc,
    Participant,
    Party,
    Pc,
    Ssc,
    Tpr,
    VoltageLevel,
)
from chellow.rate_server import download
from chellow.utils import (
    ct_datetime,
    to_ct,
    to_utc,
)


def parse_date(date_str):
    if len(date_str) == 0:
        return None
    else:
        return to_utc(to_ct(Datetime.strptime(date_str, "%d/%m/%Y")))


def parse_to_date(date_str):
    if len(date_str) == 0:
        return None
    else:
        dt = to_ct(Datetime.strptime(date_str, "%d/%m/%Y"))
        dt += Timedelta(hours=23, minutes=30)
        return to_utc(dt)


def parse_bool(bool_str):
    return bool_str == "T"


VOLTAGE_MAP = {"24": {"602": {to_utc(ct_datetime(2010, 4, 1)): "LV"}}}


def _import_Clock_Interval(sess, rows, ctx):
    cis = {
        (
            c.tpr.code,
            c.day_of_week,
            c.start_day,
            c.start_month,
            c.end_day,
            c.end_month,
            c.start_hour,
            c.start_minute,
            c.end_hour,
            c.end_minute,
        )
        for c in sess.execute(select(ClockInterval)).scalars()
    }
    for values in rows:
        tpr_code = values[0].zfill(5)
        day_of_week = int(values[1])
        start_day = int(values[2])
        start_month = int(values[3])
        end_day = int(values[4])
        end_month = int(values[5])
        start_hour, start_minute = [int(v) for v in values[6].split(":")]
        end_hour, end_minute = [int(v) for v in values[7].split(":")]

        if (
            tpr_code,
            day_of_week,
            start_day,
            start_month,
            end_day,
            end_month,
            start_hour,
            start_minute,
            end_hour,
            end_minute,
        ) not in cis:
            tpr = Tpr.get_by_code(sess, tpr_code)
            tpr.insert_clock_interval(
                sess,
                day_of_week,
                start_day,
                start_month,
                end_day,
                end_month,
                start_hour,
                start_minute,
                end_hour,
                end_minute,
            )


def _import_GSP_Group(sess, rows, ctx):
    for values in rows:
        code = values[0]
        description = values[1]

        group = GspGroup.find_by_code(sess, code)

        if group is None:
            GspGroup.insert(sess, code, description)

        else:
            group.description = description
            sess.flush()


def _import_Line_Loss_Factor_Class(sess, rows, ctx):
    VOLTAGE_LEVELS = {v.code: v for v in sess.execute(select(VoltageLevel)).scalars()}
    DNO_MAP = {
        dno.participant.code: dno
        for dno in sess.execute(
            select(Party)
            .join(MarketRole)
            .where(MarketRole.code == "R")
            .options(joinedload(Party.participant))
        ).scalars()
    }
    llfcs = {
        (v.dno.id, v.code, v.valid_from): v
        for v in sess.execute(select(Llfc)).scalars()
    }

    for values in rows:
        participant_code = values[0]
        # market_role_code = values[1]
        llfc_code = values[3].zfill(3)
        valid_from = parse_date(values[4])
        description = values[5]
        is_import = values[6] in ("A", "B")
        is_substation = any(
            p in description for p in ("_SS", " SS", " S/S", "(S/S)", "sub", "Sub")
        )

        valid_to = parse_to_date(values[7])

        try:
            dno = DNO_MAP[participant_code]
        except KeyError:
            raise BadRequest(
                f"There is no DNO with participant code {participant_code}"
            )

        try:
            voltage_level_code = VOLTAGE_MAP[dno.dno_code][llfc_code][valid_from]
        except KeyError:
            voltage_level_code = "LV"
            description_upper = description.upper()
            for vl_code in VOLTAGE_LEVELS.keys():
                if vl_code in description_upper:
                    voltage_level_code = vl_code
                    break

        voltage_level = VOLTAGE_LEVELS[voltage_level_code]

        try:
            llfc = llfcs[(dno.id, llfc_code, valid_from)]
            llfc.description = description
            llfc.voltage_level = voltage_level
            llfc.is_substation = is_substation
            llfc.is_import = is_import
            llfc.valid_to = valid_to
            sess.flush()
        except KeyError:
            dno.insert_llfc(
                sess,
                llfc_code,
                description,
                voltage_level,
                is_substation,
                is_import,
                valid_from,
                valid_to,
            )


def _import_Market_Participant(sess, rows, ctx):
    for values in rows:
        participant_code = values[0]
        participant_name = values[1]

        participant = Participant.find_by_code(sess, participant_code)

        if participant is None:
            Participant.insert(sess, participant_code, participant_name)

        else:
            participant.update(participant_name)
            sess.flush()


def _import_Market_Role(sess, rows, ctx):
    for values in rows:
        role_code = values[0]
        role_description = values[1]

        role = MarketRole.find_by_code(sess, role_code)

        if role is None:
            MarketRole.insert(sess, role_code, role_description)

        else:
            role.description = role_description
            sess.flush()


def _import_Market_Participant_Role(sess, rows, ctx):
    for values in rows:
        participant_code = values[0]
        participant = Participant.get_by_code(sess, participant_code)
        market_role_code = values[1]
        market_role = MarketRole.get_by_code(sess, market_role_code)
        valid_from = parse_date(values[2])
        party = Party.find_by_participant_role(
            sess, participant, market_role, valid_from
        )
        valid_to = parse_to_date(values[3])
        name = values[4]
        dno_code_str = values[14]
        dno_code = None if len(dno_code_str) == 0 else dno_code_str
        if dno_code == "99":
            continue

        if party is None:
            participant.insert_party(
                sess,
                market_role,
                name,
                valid_from,
                valid_to,
                dno_code,
            )

            if dno_code is not None:
                contract = Contract.find_dno_by_name(sess, dno_code)
                if contract is None:
                    Contract.insert_dno(
                        sess, dno_code, participant, "", {}, valid_from, None, {}
                    )

        else:
            party.name = name
            party.valid_to = valid_to
            party.dno_code = dno_code
            sess.flush()


def _import_Measurement_Requirement(sess, rows, ctx):
    mrs = {
        (mr.ssc.code, mr.tpr.code)
        for mr in sess.execute(select(MeasurementRequirement)).scalars()
    }
    for values in rows:
        ssc_code = Ssc.normalise_code(values[0])
        tpr_code = Tpr.normalise_code(values[1])

        if (ssc_code, tpr_code) not in mrs:
            ssc = sess.execute(
                select(Ssc).where(Ssc.code == ssc_code).order_by(Ssc.valid_from.desc())
            ).scalar()
            tpr = Tpr.get_by_code(sess, tpr_code)
            MeasurementRequirement.insert(sess, ssc, tpr)


def _import_Meter_Timeswitch_Class(sess, rows, ctx):
    ctx_mtcs = ctx["mtcs"] = {}
    meter_types = dict((m.code, m) for m in sess.execute(select(MeterType)).scalars())
    meter_payment_types = dict(
        (m.code, m) for m in sess.execute(select(MeterPaymentType)).scalars()
    )

    for values in rows:
        code = values[0].zfill(3)  # Meter Timeswitch Class ID
        valid_from = parse_date(values[1])  # Effective From Settlement Date (MTC)
        valid_to = parse_to_date(values[2])  # Effective To Settlement Date (MTC)
        description = values[3]  # Meter Timeswitch Class Description
        is_common = parse_bool(values[4])  # MTC Common Code Indicator
        has_related_metering_str = values[5]  # MTC Related Metering System Indicator
        has_related_metering = parse_bool(has_related_metering_str)

        ctx_mtc = {
            "code": code,
            "valid_from": valid_from,
            "valid_to": valid_to,
            "description": description,
            "is_common": is_common,
            "has_related_metering": has_related_metering,
        }

        if is_common:
            meter_type_code = values[6]  # Mtc Meter Type ID
            meter_type = meter_types[meter_type_code]
            meter_payment_type_code = values[7]  # MTC Payment Type ID
            meter_payment_type = meter_payment_types[meter_payment_type_code]
            has_comms_str = values[8]  # MTC Communication Indicator
            has_comms = parse_bool(has_comms_str)
            is_hh_str = values[9]  # MTC Type Indicator
            is_hh = parse_bool(is_hh_str)
            tpr_count_str = values[10]  # TPR Count
            tpr_count = None if tpr_count_str == "" else int(tpr_count_str)

            ctx_mtc["meter_type"] = meter_type
            ctx_mtc["meter_payment_type"] = meter_payment_type
            ctx_mtc["has_comms"] = has_comms
            ctx_mtc["is_hh"] = is_hh
            ctx_mtc["tpr_count"] = tpr_count

        ctx_mtcs[(code, valid_from)] = ctx_mtc

        mtc = Mtc.find_by_code(sess, code, valid_from)
        if mtc is None:
            Mtc.insert(
                sess,
                code,
                is_common,
                has_related_metering,
                valid_from,
                valid_to,
            )

        else:
            mtc.description = description
            mtc.has_related_metering = has_related_metering
            mtc.valid_to = valid_to
            sess.flush()


def _import_MTC_in_PES_Area(sess, rows, ctx):
    mtc_participants = {
        (m.participant.id, m.mtc.id, m.valid_from): m
        for m in sess.execute(
            select(MtcParticipant).options(
                joinedload(MtcParticipant.meter_type),
                joinedload(MtcParticipant.meter_payment_type),
            )
        ).scalars()
    }
    mtcs = {(m.code, m.valid_from): m for m in sess.execute(select(Mtc)).scalars()}
    participants = {p.code: p for p in sess.execute(select(Participant)).scalars()}
    meter_types = {m.code: m for m in sess.execute(select(MeterType)).scalars()}
    meter_payment_types = {
        m.code: m for m in sess.execute(select(MeterPaymentType)).scalars()
    }

    for values in rows:
        mtc_code_str = values[0]  # Meter Timeswitch Class ID
        mtc_code = mtc_code_str.zfill(3)
        mtc_from_str = values[1]  # Effective From Settlement Date (MTC)
        mtc_from = parse_date(mtc_from_str)
        mtc = mtcs[(mtc_code, mtc_from)]
        participant_code = values[2]  # Market Participant ID
        participant = participants[participant_code]
        valid_from_str = values[3]  # Effective From Settlement Date (MTCPA)
        valid_from = parse_date(valid_from_str)
        mtc_participant = mtc_participants.get((participant.id, mtc.id, valid_from))
        valid_to_str = values[4]  # Effective To Settlement Date (MTCPA)
        valid_to = parse_to_date(valid_to_str)

        if mtc.is_common:
            ctx_mtc = ctx["mtcs"][(mtc_code, mtc_from)]
            description = ctx_mtc["description"]
            meter_type = ctx_mtc["meter_type"]
            meter_payment_type = ctx_mtc["meter_payment_type"]
            has_comms = ctx_mtc["has_comms"]
            is_hh = ctx_mtc["is_hh"]
            tpr_count = ctx_mtc["tpr_count"]

        else:
            description = values[5]  # Meter Timeswitch Class Description
            meter_type_code = values[6]  # Mtc Meter Type ID
            meter_type = meter_types[meter_type_code]
            meter_payment_type_code = values[7]  # MTC Payment Type ID
            meter_payment_type = meter_payment_types[meter_payment_type_code]
            has_comms = values[8] == "Y"  # MTC Comm. Indicator
            is_hh = values[9] == "H"  # MTC Type Indicator
            tpr_count_str = values[10]  # TPR Count
            tpr_count = 0 if tpr_count_str == "" else int(tpr_count_str)

        if mtc_participant is None:
            MtcParticipant.insert(
                sess,
                mtc,
                participant,
                description,
                has_comms,
                is_hh,
                meter_type,
                meter_payment_type,
                tpr_count,
                valid_from,
                valid_to,
            )

        else:
            mtc.description = description
            mtc.has_comms = has_comms
            mtc.is_hh = is_hh
            mtc.meter_type = meter_type
            mtc.meter_payment_type = meter_payment_type
            mtc.tpr_count = tpr_count
            mtc.valid_to = valid_to
            sess.flush()


def _import_MTC_Meter_Type(sess, rows, ctx):
    meter_types = dict(
        ((v.code, v.valid_from), v) for v in sess.execute(select(MeterType)).scalars()
    )
    for values in rows:
        code = values[0]
        description = values[1]
        valid_from = parse_date(values[2])
        valid_to = parse_to_date(values[3])
        meter_type = meter_types.get((code, valid_from))

        if meter_type is None:
            MeterType.insert(sess, code, description, valid_from, valid_to)

        else:
            meter_type.description = description
            meter_type.valid_to = valid_to
            sess.flush()


def _import_MTC_Payment_Type(sess, rows, ctx):
    meter_payment_types = dict(
        ((v.code, v.valid_from), v)
        for v in sess.execute(select(MeterPaymentType)).scalars()
    )
    for values in rows:
        code = values[0]
        description = values[1]
        valid_from = parse_date(values[2])
        valid_to = parse_to_date(values[3])
        meter_payment_type = meter_payment_types.get((code, valid_from))

        if meter_payment_type is None:
            MeterPaymentType.insert(sess, code, description, valid_from, valid_to)

        else:
            meter_payment_type.description = description
            meter_payment_type.valid_to = valid_to
            sess.flush()


def _import_Profile_Class(sess, rows, ctx):
    for values in rows:
        code_str = values[0]
        code = code_str.zfill(2)
        valid_from = parse_date(values[1])
        name = values[2]
        valid_to = parse_to_date(values[4])

        pc = Pc.find_by_code(sess, code)

        if pc is None:
            Pc.insert(sess, code, name, valid_from, valid_to)

        else:
            pc.name = name
            pc.valid_to = valid_to
            sess.flush()


def _import_Standard_Settlement_Configuration(sess, rows, ctx):
    sscs = dict(
        ((v.code, v.valid_from), v) for v in sess.execute(select(Ssc)).scalars()
    )
    for values in rows:
        code = values[0]
        valid_from = parse_date(values[1])
        valid_to = parse_to_date(values[2])
        description = values[3]
        is_import = values[4] == "I"
        ssc = sscs.get((code, valid_from))

        if ssc is None:
            Ssc.insert(sess, code, description, is_import, valid_from, valid_to)

        else:
            ssc.description = description
            ssc.is_import = is_import
            ssc.valid_to = valid_to
            sess.flush()


def _import_Time_Pattern_Regime(sess, rows, ctx):
    for values in rows:
        code = values[0]
        is_teleswitch_str = values[1]
        is_teleswitch = is_teleswitch_str == "S"
        is_gmt_str = values[2]
        is_gmt = is_gmt_str == "Y"

        tpr = Tpr.find_by_code(sess, code)

        if tpr is None:
            Tpr.insert(sess, code, is_teleswitch, is_gmt)

        else:
            tpr.is_teleswitch = is_teleswitch
            tpr.is_gmt = is_gmt
            sess.flush()


def _import_Valid_MTC_LLFC_Combination(sess, rows, ctx):
    mtcs = {(v.code, v.valid_from): v for v in sess.execute(select(Mtc)).scalars()}
    participants = {v.code: v for v in sess.execute(select(Participant)).scalars()}
    mtc_participants = {
        (v.mtc.id, v.participant.id, v.valid_from): v
        for v in sess.execute(select(MtcParticipant)).scalars()
    }
    mtc_llfcs = {
        (v.mtc_participant.id, v.llfc.id, v.valid_from): v
        for v in sess.execute(select(MtcLlfc)).scalars()
    }
    dnos = {}
    llfcs = {}

    for values in rows:
        mtc_code_str = values[0]  # Meter Timeswitch Class ID
        mtc_code = mtc_code_str.zfill(3)
        mtc_from_str = values[1]  # Effective From Settlement Date (MTC)
        mtc_from = parse_date(mtc_from_str)
        mtc = mtcs[(mtc_code, mtc_from)]
        participant_code = values[2]  # Market Participant ID
        participant = participants[participant_code]

        mtc_participant_from_str = values[3]  # Effective From Settlement Date (MTCPA)
        mtc_participant_from = parse_date(mtc_participant_from_str)
        mtc_participant = mtc_participants[
            (mtc.id, participant.id, mtc_participant_from)
        ]
        llfc_code = values[4]  # Line Loss Factor Class ID
        valid_from_str = values[5]  # Effective From Settlement Date (VMTCLC)
        valid_from = parse_date(valid_from_str)
        valid_to_str = values[6]  # Effective To Settlement Date (VMTCLC)
        valid_to = parse_to_date(valid_to_str)

        try:
            dno = dnos[(participant.id, valid_from)]
        except KeyError:
            dno = dnos[(participant.id, valid_from)] = sess.execute(
                select(Party)
                .join(Participant)
                .join(MarketRole)
                .where(
                    Participant.code == participant_code,
                    MarketRole.code == "R",
                    Party.valid_from <= valid_from,
                    or_(Party.valid_to == null(), Party.valid_to >= valid_from),
                )
            ).scalar_one()
        try:
            llfc = llfcs[(dno.id, llfc_code, valid_from)]
        except KeyError:
            llfc = llfcs[(dno.id, llfc_code, valid_from)] = dno.get_llfc_by_code(
                sess, llfc_code, valid_from
            )

        try:
            mtc_llfc = mtc_llfcs[(mtc_participant.id, llfc.id, valid_from)]
            mtc_llfc.valid_to = valid_to
            sess.flush()
        except KeyError:
            MtcLlfc.insert(
                sess,
                mtc_participant,
                llfc,
                valid_from,
                valid_to,
            )


def _import_Valid_MTC_SSC_Combination(sess, rows, ctx):
    mtcs = dict(
        ((v.code, v.valid_from), v) for v in sess.execute(select(Mtc)).scalars()
    )
    participants = dict(
        (v.code, v) for v in sess.execute(select(Participant)).scalars()
    )
    mtc_participants = dict(
        ((v.mtc.id, v.participant.id, v.valid_from), v)
        for v in sess.execute(select(MtcParticipant)).scalars()
    )
    sscs = dict(
        ((v.code, v.valid_from), v) for v in sess.execute(select(Ssc)).scalars()
    )
    mtc_sscs = dict(
        ((v.mtc_participant.id, v.ssc.id, v.valid_from), v)
        for v in sess.execute(select(MtcSsc)).scalars()
    )

    for values in rows:
        mtc_code = values[0].zfill(3)  # Meter Timeswitch Class ID
        mtc_from_str = values[1]  # Effective From Settlement Date (MTC)
        mtc_from = parse_date(mtc_from_str)
        mtc = mtcs[(mtc_code, mtc_from)]
        participant_code = values[2]  # Market Participant ID
        participant = participants[participant_code]
        mtc_participant_from_str = values[3]  # Effective From Settlement Date (MTCPA)
        mtc_participant_from = parse_date(mtc_participant_from_str)
        mtc_participant = mtc_participants[
            (mtc.id, participant.id, mtc_participant_from)
        ]
        ssc_code = values[4]  # Standard Settlement Configuration ID
        valid_from_str = values[5]  # Effective From Settlement Date (VMTCSC)
        valid_from = parse_date(valid_from_str)
        valid_to_str = values[6]  # Effective To Settlement Date (VMTCSC)
        valid_to = parse_to_date(valid_to_str)

        try:
            ssc = sscs[(ssc_code, valid_from)]
        except KeyError:
            ssc = sscs[(ssc_code, valid_from)] = Ssc.get_by_code(
                sess, ssc_code, valid_from
            )

        try:
            mtc_ssc = mtc_sscs[(mtc_participant.id, ssc.id, valid_from)]
            mtc_ssc.valid_to = valid_to
            sess.flush()
        except KeyError:
            MtcSsc.insert(
                sess,
                mtc_participant,
                ssc,
                valid_from,
                valid_to,
            )


def _import_Valid_MTC_LLFC_SSC_Combination(sess, rows, ctx):
    mtcs = {(v.code, v.valid_from): v for v in sess.execute(select(Mtc)).scalars()}
    participants = {v.code: v for v in sess.execute(select(Participant)).scalars()}
    mtc_participants = {
        (v.mtc.id, v.participant.id, v.valid_from): v
        for v in sess.execute(select(MtcParticipant)).scalars()
    }
    sscs = {(v.code, v.valid_from): v for v in sess.execute(select(Ssc)).scalars()}
    mtc_sscs = {
        (v.mtc_participant.id, v.ssc.id, v.valid_from): v
        for v in sess.execute(select(MtcSsc)).scalars()
    }
    dnos = {}
    llfcs = {}
    mtc_llfc_sscs = {
        (v.mtc_ssc.id, v.llfc.id, v.valid_from): v
        for v in sess.execute(select(MtcLlfcSsc)).scalars()
    }

    for values in rows:
        mtc_code = values[0].zfill(3)  # Meter Timeswitch Class ID
        mtc_from_str = values[1]  # Effective From Settlement Date (MTC)
        mtc_from = parse_date(mtc_from_str)
        mtc = mtcs[(mtc_code, mtc_from)]
        participant_code = values[2]  # Market Participant ID
        participant = participants[participant_code]
        mtc_participant_from_str = values[3]  # Effective From Settlement Date (MTCPA)
        mtc_participant_from = parse_date(mtc_participant_from_str)
        mtc_participant = mtc_participants[
            (mtc.id, participant.id, mtc_participant_from)
        ]
        ssc_code = values[4]  # Standard Settlement Configuration ID
        mtc_ssc_from_str = values[5]  # Effective From Settlement date (VMTCSC)
        mtc_ssc_from = parse_date(mtc_ssc_from_str)
        try:
            ssc = sscs[(ssc_code, mtc_ssc_from)]
        except KeyError:
            ssc = sscs[(ssc_code, mtc_ssc_from)] = Ssc.get_by_code(
                sess, ssc_code, mtc_ssc_from
            )
        mtc_ssc = mtc_sscs[(mtc_participant.id, ssc.id, mtc_ssc_from)]
        llfc_code = values[6]  # Line Loss Factor Class ID
        valid_from_str = values[7]  # Effective From Settlement Date (VMTCLSC)
        valid_from = parse_date(valid_from_str)
        valid_to_str = values[8]  # Effective To Settlement Date (VMTCLSC)
        valid_to = parse_to_date(valid_to_str)

        try:
            dno = dnos[(participant.id, valid_from)]
        except KeyError:
            dno = dnos[(participant.id, valid_from)] = sess.execute(
                select(Party)
                .join(Participant)
                .join(MarketRole)
                .where(
                    Participant.code == participant_code,
                    MarketRole.code == "R",
                    Party.valid_from <= valid_from,
                    or_(Party.valid_to == null(), Party.valid_to >= valid_from),
                )
            ).scalar_one()

        try:
            llfc = llfcs[(dno.id, llfc_code, valid_from)]
        except KeyError:
            llfc = llfcs[(dno.id, llfc_code, valid_from)] = dno.get_llfc_by_code(
                sess, llfc_code, valid_from
            )

        try:
            mtc_llfc_ssc = mtc_llfc_sscs[mtc_ssc.id, llfc.id, valid_from]
            mtc_llfc_ssc.valid_to = valid_to
            sess.flush()
        except KeyError:
            MtcLlfcSsc.insert(
                sess,
                mtc_ssc,
                llfc,
                valid_from,
                valid_to,
            )


def _import_Valid_MTC_LLFC_SSC_PC_Combination(sess, rows, ctx):
    mtcs = {(v.code, v.valid_from): v for v in sess.execute(select(Mtc)).scalars()}
    participants = {v.code: v for v in sess.execute(select(Participant)).scalars()}
    mtc_participants = {
        (v.mtc.id, v.participant.id, v.valid_from): v
        for v in sess.execute(select(MtcParticipant)).scalars()
    }
    mtc_sscs = {
        (v.mtc_participant.id, v.ssc.id, v.valid_from): v
        for v in sess.execute(select(MtcSsc)).scalars()
    }
    mtc_llfc_sscs = {
        (v.mtc_ssc.id, v.llfc.id, v.valid_from): v
        for v in sess.execute(select(MtcLlfcSsc)).scalars()
    }
    dnos = {}
    llfcs = {}
    sscs = {}
    pcs = {v.code: v for v in sess.execute(select(Pc)).scalars()}
    combos = {
        (v.mtc_llfc_ssc.id, v.pc.id, v.valid_from): v
        for v in sess.execute(select(MtcLlfcSscPc)).scalars()
    }
    for values in rows:
        mtc_code = values[0].zfill(3)  # Meter Timeswitch Class ID
        mtc_from_str = values[1]  # Effective From Settlement Date (MTC)
        mtc_from = parse_date(mtc_from_str)
        mtc = mtcs[(mtc_code, mtc_from)]
        participant_code = values[2]  # Market Participant ID
        participant = participants[participant_code]
        mtc_participant_from_str = values[3]  # Effective From Settlement Date (MTCPA)
        mtc_participant_from = parse_date(mtc_participant_from_str)
        mtc_participant = mtc_participants[
            (mtc.id, participant.id, mtc_participant_from)
        ]
        ssc_code = values[4]  # Standard Settlement Configuration ID
        mtc_ssc_from_str = values[5]  # Effective From Settlement date (VMTCSC)
        mtc_ssc_from = parse_date(mtc_ssc_from_str)

        try:
            ssc = sscs[(ssc_code, mtc_ssc_from)]
        except KeyError:
            ssc = sscs[(ssc_code, mtc_ssc_from)] = Ssc.get_by_code(
                sess, ssc_code, mtc_ssc_from
            )

        mtc_ssc = mtc_sscs[(mtc_participant.id, ssc.id, mtc_ssc_from)]
        llfc_code = values[6]  # Line Loss Factor Class ID
        mtc_llfc_ssc_from_str = values[7]  # Effective From Settlement Date (VMTCLSC)
        mtc_llfc_ssc_from = parse_date(mtc_llfc_ssc_from_str)
        pc_code = values[8].zfill(2)  # Profile Class ID
        valid_from_str = values[9]  # Effective From Settlement Date (VMTCLSPC)
        valid_from = parse_date(valid_from_str)
        valid_to_str = values[10]  # Effective To Settlement Date (VMTCLSPC)
        valid_to = parse_to_date(valid_to_str)

        try:
            dno = dnos[(participant_code, valid_from)]
        except KeyError:
            dno = dnos[(participant_code, valid_from)] = sess.execute(
                select(Party)
                .join(Participant)
                .join(MarketRole)
                .where(
                    Participant.code == participant_code,
                    MarketRole.code == "R",
                    Party.valid_from <= valid_from,
                    or_(Party.valid_to == null(), Party.valid_to >= valid_from),
                )
            ).scalar_one()

        try:
            llfc = llfcs[(dno.id, llfc_code, mtc_llfc_ssc_from)]
        except KeyError:
            llfc = llfcs[(dno.id, llfc_code, mtc_llfc_ssc_from)] = dno.get_llfc_by_code(
                sess, llfc_code, mtc_llfc_ssc_from
            )

        mtc_llfc_ssc = mtc_llfc_sscs[(mtc_ssc.id, llfc.id, mtc_llfc_ssc_from)]

        pc = pcs[pc_code]
        try:
            combo = combos[(mtc_llfc_ssc.id, pc.id, valid_from)]
            combo.valid_to = valid_to
            sess.flush()
        except KeyError:
            MtcLlfcSscPc.insert(
                sess,
                mtc_llfc_ssc,
                pc,
                valid_from,
                valid_to,
            )


def rate_server_import(sess, log, set_progress, s, paths):
    log("Starting to check for a new MDD version")
    mdd_entries = {}
    for path, url in paths:
        if len(path) == 5:
            _, utility, rate_type, mdd_version_str, file_name = path

            if utility == "electricity" and rate_type == "mdd":
                mdd_version = int(mdd_version_str)
                try:
                    fl_entries = mdd_entries[mdd_version]
                except KeyError:
                    fl_entries = mdd_entries[mdd_version] = {}

                fl_entries[file_name] = url

    if len(mdd_entries) == 0:
        raise BadRequest("Can't find any MDD versions on the rate server.")

    mdd_version, fl_entries = sorted(mdd_entries.items())[-1]
    log(f"Latest version on rate server: {mdd_version}.")

    config = Contract.get_non_core_by_name(sess, "configuration")
    state = config.make_state()
    current_version = state.get("mdd_version", 0)

    log(f"Latest version in Chellow: {current_version}")
    if mdd_version <= current_version:
        return

    gnames = {}
    ctx = {}
    version = None

    for file_name, url in fl_entries.items():
        csv_file = StringIO(download(s, url).decode("utf8"))
        csv_reader = iter(csv.reader(csv_file))
        next(csv_reader)  # Skip titles

        table_name_elements = file_name.split("_")
        ver = int(table_name_elements[-1].split(".")[0])
        if version is None:
            version = ver

        if version != ver:
            raise BadRequest(
                f"There's a mixture of MDD versions in the file names. Expected "
                f"version {version} but found version {ver} in {file_name}."
            )
        table_name = "_".join(table_name_elements[:-1])
        gnames[table_name] = list(csv_reader)

    for tname, func in [
        ("GSP_Group", _import_GSP_Group),
        ("Market_Participant", _import_Market_Participant),
        ("Market_Role", _import_Market_Role),
        ("Market_Participant_Role", _import_Market_Participant_Role),
        ("Line_Loss_Factor_Class", _import_Line_Loss_Factor_Class),
        ("MTC_Meter_Type", _import_MTC_Meter_Type),
        ("MTC_Payment_Type", _import_MTC_Payment_Type),
        ("Meter_Timeswitch_Class", _import_Meter_Timeswitch_Class),
        ("MTC_in_PES_Area", _import_MTC_in_PES_Area),
        (
            "Standard_Settlement_Configuration",
            _import_Standard_Settlement_Configuration,
        ),
        (
            "Valid_MTC_LLFC_Combination",
            _import_Valid_MTC_LLFC_Combination,
        ),
        (
            "Valid_MTC_SSC_Combination",
            _import_Valid_MTC_SSC_Combination,
        ),
        (
            "Valid_MTC_LLFC_SSC_Combination",
            _import_Valid_MTC_LLFC_SSC_Combination,
        ),
        ("Profile_Class", _import_Profile_Class),
        (
            "Valid_MTC_LLFC_SSC_PC_Combination",
            _import_Valid_MTC_LLFC_SSC_PC_Combination,
        ),
        ("Time_Pattern_Regime", _import_Time_Pattern_Regime),
        ("Clock_Interval", _import_Clock_Interval),
        ("Measurement_Requirement", _import_Measurement_Requirement),
    ]:
        if tname in gnames:
            log(f"Found {tname} and will now import it.")
            func(sess, gnames[tname], ctx)
        else:
            raise BadRequest(f"Can't find {tname} on the rate server.")

    config = Contract.get_non_core_by_name(sess, "configuration")
    state = config.make_state()
    state["mdd_version"] = version
    config.update_state(state)
    sess.commit()



import pytest

from chellow.e.bsuos import _find_file_type, _process_url


@pytest.mark.parametrize(
    "filename,ext",
    [
        ['filename="bsuos.csv"', "csv"],
        ['inline; filename="Current_RF_BSUoS_Data_175.xls"', "xls"],
        ["inline; filename=Current_RF_BSUoS_Data_175.xls", "xls"],
    ],
)
def test_find_file_type_csv(filename, ext):
    assert _find_file_type(filename) == ext


def test_process_url_csv(mocker, sess):
    res = mocker.Mock()
    res.text = "Date"
    res.status_code = 200
    res.reason = "okay"
    res.headers = {"Content-Disposition": 'filename="bsuos.csv"'}
    mocker.patch("chellow.e.bsuos.requests.get", return_value=res)
    logger = mocker.Mock()
    contract = mocker.Mock()
    url = mocker.Mock()
    _process_url(logger, sess, url, contract)



from decimal import Decimal, InvalidOperation
from io import BytesIO

from dateutil.relativedelta import relativedelta

from openpyxl import load_workbook

from sqlalchemy import null, or_, select

from werkzeug.exceptions import BadRequest

import chellow.e.computer
import chellow.e.duos
from chellow.models import Contract, RateScript
from chellow.rate_server import download
from chellow.utils import (
    c_months_u,
    ct_datetime,
    hh_after,
    hh_format,
    hh_min,
    to_ct,
    to_utc,
)

BANDED_START = to_utc(ct_datetime(2023, 4, 1))


def hh(ds, rate_period="monthly", est_kw=None):
    for hh in ds.hh_data:
        if hh["start-date"] >= BANDED_START:
            if hh["ct-decimal-hour"] == 12:
                _process_banded_hh(ds, hh)
        else:
            if hh["ct-is-month-end"]:
                _process_triad_hh(ds, rate_period, est_kw, hh)


def _process_banded_hh(ds, hh):
    rates = ds.non_core_rate("tnuos", hh["start-date"])
    lookup = rates["lookup"]
    band = lookup[hh["duos-description"]]
    hh["tnuos-band"] = band
    rate = float(rates["bands"][band])
    if band == "Unmetered":
        hh["tnuos-gbp"] = rate / 100 * ds.sc / 365
    else:
        hh["tnuos-gbp"] = rate


def _process_triad_hh(ds, rate_period, est_kw, hh):
    month_start, month_finish = next(
        c_months_u(start_year=hh["ct-year"], start_month=hh["ct-month"])
    )

    month_start_ct = to_ct(month_start)
    if month_start_ct.month > 3:
        year = month_start_ct.year
    else:
        year = month_start_ct.year - 1
    financial_year_start = to_utc(ct_datetime(year, 4, 1))
    last_financial_year_start = to_utc(ct_datetime(year - 1, 4, 1))
    financial_year_finish = to_utc(ct_datetime(year + 1, 3, 31, 23, 30))

    est_triad_kws = []
    earliest_triad = None
    for dt in ds.non_core_rate("triad_dates", last_financial_year_start)["triad_dates"]:
        triad_hh = None
        earliest_triad = hh_min(earliest_triad, dt)
        try:
            d = next(ds.get_data_sources(dt, dt, financial_year_start))
            chellow.e.duos.duos_vb(d)
            triad_hh = d.hh_data[0]

            while dt < financial_year_start:
                dt += relativedelta(years=1)

            for d in ds.get_data_sources(dt, dt, financial_year_start):
                chellow.e.duos.duos_vb(d)
                datum = d.hh_data[0]
                triad_hh["laf"] = datum["laf"]
                triad_hh["gsp-kw"] = datum["laf"] * triad_hh["msp-kw"]
        except StopIteration:
            triad_hh = {
                "hist-start": dt,
                "msp-kw": 0,
                "start-date": dt,
                "status": "before start of MPAN",
                "laf": 1,
                "gsp-kw": 0,
            }
        est_triad_kws.append(triad_hh)

    if ds.site is None:
        era = ds.supply.find_era_at(ds.sess, earliest_triad)
        if (
            era is None
            or era.get_channel(ds.sess, ds.is_import, "ACTIVE") is None
            and est_kw is None
        ):
            est_kw = 0.85 * max(datum["msp-kwh"] for datum in ds.hh_data) * 2
        if est_kw is not None:
            for est_datum in est_triad_kws:
                est_datum["msp-kw"] = est_kw
                est_datum["gsp-kw"] = est_datum["msp-kw"] * est_datum["laf"]

    gsp_kw = 0
    for i, triad_hh in enumerate(est_triad_kws):
        triad_prefix = "triad-estimate-" + str(i + 1)
        hh[triad_prefix + "-date"] = triad_hh["hist-start"]
        hh[triad_prefix + "-msp-kw"] = triad_hh["msp-kw"]
        hh[triad_prefix + "-status"] = triad_hh["status"]
        hh[triad_prefix + "-laf"] = triad_hh["laf"]
        hh[triad_prefix + "-gsp-kw"] = triad_hh["gsp-kw"]
        gsp_kw += triad_hh["gsp-kw"]

    hh["triad-estimate-gsp-kw"] = gsp_kw / 3
    polarity = "import" if ds.llfc.is_import else "export"
    gsp_group_code = ds.gsp_group_code
    rate = float(
        ds.non_core_rate("tnuos", month_start)["triad_gbp_per_gsp_kw"][polarity][
            gsp_group_code
        ]
    )

    hh["triad-estimate-rate"] = rate

    est_triad_gbp = hh["triad-estimate-rate"] * hh["triad-estimate-gsp-kw"]

    if rate_period == "monthly":
        total_intervals = 12

        est_intervals = 1
        hh["triad-estimate-months"] = est_intervals
    else:
        dt = financial_year_start
        total_intervals = 0
        while dt <= financial_year_finish:
            total_intervals += 1
            dt += relativedelta(days=1)

        est_intervals = 0
        for d in ds.get_data_sources(month_start, month_finish):
            for h in d.hh_data:
                if h["ct-decimal-hour"] == 0:
                    est_intervals += 1

        hh["triad-estimate-days"] = est_intervals

    hh["triad-estimate-gbp"] = est_triad_gbp / total_intervals * est_intervals

    if hh["ct-month"] == 3:
        triad_kws = []
        for t_date in ds.non_core_rate("triad_dates", month_start)["triad_dates"]:
            try:
                d = next(ds.get_data_sources(t_date, t_date))
                if (
                    ds.supplier_contract is None
                    or d.supplier_contract == ds.supplier_contract
                ):
                    chellow.e.duos.duos_vb(d)
                    thh = d.hh_data[0]
                else:
                    thh = {
                        "hist-start": t_date,
                        "msp-kw": 0,
                        "start-date": t_date,
                        "status": "before contract",
                        "laf": "before contract",
                        "gsp-kw": 0,
                    }
            except StopIteration:
                thh = {
                    "hist-start": t_date,
                    "msp-kw": 0,
                    "start-date": t_date,
                    "status": "before start of supply",
                    "laf": "before start of supply",
                    "gsp-kw": 0,
                }

            while t_date < financial_year_start:
                t_date += relativedelta(years=1)

            try:
                d = next(ds.get_data_sources(t_date, t_date))
                if (
                    ds.supplier_contract is None
                    or d.supplier_contract == ds.supplier_contract
                ):
                    chellow.e.duos.duos_vb(d)
                    thh["laf"] = d.hh_data[0]["laf"]
                    thh["gsp-kw"] = thh["laf"] * thh["msp-kw"]
            except StopIteration:
                pass

            triad_kws.append(thh)
        gsp_kw = 0

        for i, triad_hh in enumerate(triad_kws):
            pref = "triad-actual-" + str(i + 1)
            hh[pref + "-date"] = triad_hh["start-date"]
            hh[pref + "-msp-kw"] = triad_hh["msp-kw"]
            hh[pref + "-status"] = triad_hh["status"]
            hh[pref + "-laf"] = triad_hh["laf"]
            hh[pref + "-gsp-kw"] = triad_hh["gsp-kw"]
            gsp_kw += triad_hh["gsp-kw"]

        hh["triad-actual-gsp-kw"] = gsp_kw / 3
        polarity = "import" if ds.llfc.is_import else "export"
        gsp_group_code = ds.gsp_group_code
        tot_rate = 0
        for rs in ds.sess.execute(
            select(RateScript)
            .join(Contract, RateScript.contract_id == Contract.id)
            .where(
                Contract.name == "tnuos",
                RateScript.start_date <= financial_year_finish,
                or_(
                    RateScript.finish_date == null(),
                    RateScript.finish_date >= financial_year_start,
                ),
            )
        ).scalars():
            start_month = to_ct(rs.start_date).month
            if start_month < 4:
                start_month += 12

            if rs.finish_date is None:
                finish_month = 3
            else:
                finish_month = to_ct(rs.finish_date).month

            if finish_month < 4:
                finish_month += 12

            rt = ds.non_core_rate("tnuos", rs.start_date)["triad_gbp_per_gsp_kw"][
                polarity
            ][gsp_group_code]
            tot_rate += (finish_month - start_month + 1) * float(rt)

        rate = tot_rate / 12
        hh["triad-actual-rate"] = rate

        hh["triad-actual-gbp"] = hh["triad-actual-rate"] * hh["triad-actual-gsp-kw"]

        era = ds.supply.find_era_at(ds.sess, month_finish)
        est_intervals = 0

        interval = (
            relativedelta(months=1)
            if rate_period == "monthly"
            else relativedelta(days=1)
        )

        dt = month_finish
        while era is not None and dt > financial_year_start:
            est_intervals += 1
            dt -= interval
            if hh_after(dt, era.finish_date):
                era = ds.supply.find_era_at(ds.sess, dt)

        if rate_period == "monthly":
            hh["triad-all-estimates-months"] = est_intervals
        else:
            hh["triad-all-estimates-days"] = est_intervals
        hh["triad-all-estimates-gbp"] = (
            est_triad_gbp / total_intervals * est_intervals * -1
        )


def get_cell(sheet, col, row):
    try:
        coordinates = f"{col}{row}"
        return sheet[coordinates]
    except IndexError:
        raise BadRequest(f"Can't find the cell {coordinates} on sheet {sheet}.")


def get_int(sheet, col, row):
    return int(get_cell(sheet, col, row).value)


def get_dec(sheet, col, row):
    cell = get_cell(sheet, col, row)
    try:
        return Decimal(str(cell.value))
    except InvalidOperation as e:
        raise BadRequest(f"Problem parsing the number at {cell.coordinate}. {e}")


def get_str(sheet, col, row):
    return get_cell(sheet, col, row).value.strip()


def get_value(row, idx):
    val = row[idx].value
    if isinstance(val, str):
        return val.strip()
    else:
        return val


def find_bands(tabs):
    tab_name = "T10"
    try:
        sheet = tabs[tab_name]
    except KeyError:
        raise BadRequest(f"Can't find the tab named '{tab_name}'")

    bands = {}

    in_bands = False
    for row in range(1, len(sheet["B"]) + 1):
        val = get_cell(sheet, "B", row).value
        val_0 = None if val is None else " ".join(val.split())
        if in_bands:
            if val_0 is None or len(val_0) == 0 or val_0 == "Demand Residual (£m)":
                in_bands = False
            elif val_0 != "Unmetered demand":
                bands[val_0] = get_dec(sheet, "E", row)

        elif val_0 == "Band":
            in_bands = True

    return bands


def find_triad(tabs):
    GSP_LOOKUP = {
        1: "_P",
        2: "_N",
        3: "_F",
        4: "_G",
        5: "_M",
        6: "_D",
        7: "_B",
        8: "_E",
        9: "_A",
        10: "_K",
        11: "_J",
        12: "_C",
        13: "_H",
        14: "_L",
    }

    tab_name = "T9"
    try:
        sheet = tabs[tab_name]
    except KeyError:
        raise BadRequest(f"Can't find the tab named '{tab_name}'")

    triads = {"import": {}, "export": {}}

    in_triads = False
    for row in range(1, len(sheet["A"]) + 1):
        val = get_cell(sheet, "A", row).value
        if val is None:
            val_0 = None
        elif isinstance(val, str):
            val_0 = " ".join(val.split())
        else:
            val_0 = val

        if in_triads:
            if val_0 is None or val_0 == "":
                in_triads = False
            else:
                gsp_group_code = GSP_LOOKUP[val_0]
                triads["import"][gsp_group_code] = get_dec(sheet, "C", row)
                triads["export"][gsp_group_code] = get_dec(sheet, "E", row)

        elif val_0 == "Zone":
            in_triads = True

    return triads


def rate_server_import(sess, log, set_progress, s, paths):
    log("Starting to check for new TNUoS spreadsheets")
    year_entries = {}
    for path, url in paths:
        if len(path) == 4:
            year_str, utility, rate_type, file_name = path
            year = int(year_str)

            if utility == "electricity" and rate_type == "tnuos":
                try:
                    fl_entries = year_entries[year]
                except KeyError:
                    fl_entries = year_entries[year] = {}

                fl_entries[file_name] = url

    contract = Contract.get_non_core_by_name(sess, "tnuos")
    for year, fl_entries in sorted(year_entries.items()):
        fy_start = to_utc(ct_datetime(year, 4, 1))
        if fy_start < contract.start_rate_script.start_date:
            continue

        rs = sess.execute(
            select(RateScript).where(
                RateScript.contract == contract,
                RateScript.start_date == fy_start,
            )
        ).scalar_one_or_none()
        if rs is None:
            rs = contract.insert_rate_script(sess, fy_start, {})

        file_name, url = sorted(fl_entries.items())[-1]

        rs_script = rs.make_script()
        if rs_script.get("a_file_name") != file_name:
            try:
                fl = BytesIO(download(s, url))
                book = load_workbook(fl, data_only=True)
                tabs = {sheet.title.strip(): sheet for sheet in book.worksheets}
                try:
                    bands = find_bands(tabs)
                    triad = find_triad(tabs)
                except BadRequest as e:
                    raise BadRequest(
                        f"Problem with file '{file_name}': {e.description}"
                    )
                rs_script["bands"] = bands
                rs_script["triad_gbp_per_gsp_kw"] = triad
                rs_script["a_file_name"] = file_name
                rs.update(rs_script)
                log(f"Updated TNUoS rate script for " f"{hh_format(fy_start)}")
            except BadRequest as e:
                raise BadRequest(f"Problem with year {year}: {e.description}")

    log("Finished TNUoS spreadsheets")
    sess.commit()



from chellow.e.computer import SupplySource
from chellow.e.tnuos import _process_banded_hh
from chellow.models import (
    Comm,
    Contract,
    Cop,
    EnergisationStatus,
    GspGroup,
    MarketRole,
    MeterPaymentType,
    MeterType,
    Mtc,
    MtcLlfc,
    MtcParticipant,
    Participant,
    Pc,
    Site,
    Source,
    VoltageLevel,
    insert_comms,
    insert_cops,
    insert_energisation_statuses,
    insert_sources,
    insert_voltage_levels,
)
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_process_banded_hh_ums(sess):
    vf = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")
    start_date = to_utc(ct_datetime(2023, 7, 31, 23, 30))
    finish_date = to_utc(ct_datetime(2023, 7, 31, 23, 30))
    forecast_from = to_utc(ct_datetime(2020, 1, 1))

    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess, "bank_holidays", "", {}, vf, None, bank_holiday_rate_script
    )
    tnuos_rate_script = {
        "lookup": {"Unmetered": "Unmetered"},
        "bands": {"Unmetered": 1},
    }
    Contract.insert_non_core(
        sess,
        "tnuos",
        "",
        {},
        vf,
        None,
        tnuos_rate_script,
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop Ltd", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, vf, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        "",
        {},
        vf,
        None,
        {},
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    Contract.insert_dno(sess, dno.dno_code, participant, "", {}, vf, None, {})
    meter_type = MeterType.insert(sess, "UM", "Unmetered", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", True, False, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "510", "PC 5-8 & HH HV", voltage_level, False, True, vf, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "net")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        to_utc(ct_datetime(2000, 1, 1)),
        None,
        gsp_group,
        mop_contract,
        "773",
        dc_contract,
        "ghyy3",
        "hgjeyhuw",
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        {},
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )

    sess.commit()
    caches = {}
    era = supply.eras[0]
    is_import = True
    ds = SupplySource(
        sess, start_date, finish_date, forecast_from, era, is_import, caches
    )
    hh = ds.hh_data[0]
    hh["duos-description"] = "Unmetered"
    _process_banded_hh(ds, hh)
    assert hh == {
        "hist-start": utc_datetime(2019, 7, 31, 22, 30),
        "start-date": utc_datetime(2023, 7, 31, 22, 30),
        "ct-day": 31,
        "utc-month": 7,
        "utc-day": 31,
        "utc-decimal-hour": 22.5,
        "utc-year": 2023,
        "utc-hour": 22,
        "utc-minute": 30,
        "ct-year": 2023,
        "ct-month": 7,
        "ct-decimal-hour": 23.5,
        "ct-day-of-week": 0,
        "utc-day-of-week": 0,
        "utc-is-bank-holiday": False,
        "ct-is-bank-holiday": False,
        "utc-is-month-end": False,
        "ct-is-month-end": True,
        "status": "X",
        "imp-msp-kvarh": 0,
        "imp-msp-kvar": 0,
        "exp-msp-kvarh": 0,
        "exp-msp-kvar": 0,
        "msp-kw": 0,
        "msp-kwh": 0,
        "hist-import-net-kvarh": 0,
        "hist-export-net-kvarh": 0,
        "anti-msp-kwh": 0,
        "anti-msp-kw": 0,
        "hist-imp-msp-kvarh": 0,
        "hist-kwh": 0,
        "duos-description": "Unmetered",
        "tnuos-band": "Unmetered",
        "tnuos-gbp": 0.00989041095890411,
    }



from flask.testing import FlaskClient

from jinja2 import Environment, PackageLoader, select_autoescape

import pg8000

import pytest

from requests.auth import _basic_auth_str


import chellow.models
from chellow import create_app
from chellow.models import Session, User, UserRole, stop_sqlalchemy


@pytest.fixture
def fresh_db():
    stop_sqlalchemy()
    config = chellow.models.config
    database = config["PGDATABASE"]
    with pg8000.connect(
        config["PGUSER"],
        host=config["PGHOST"],
        database="postgres",
        port=int(config["PGPORT"]),
        password=config["PGPASSWORD"],
    ) as con:
        cursor = con.cursor()
        con.rollback()
        con.autocommit = True
        cursor.execute(f"DROP DATABASE IF EXISTS {database};")
        cursor.execute(f"CREATE DATABASE {database} ENCODING 'UTF8';")


@pytest.fixture
def app(fresh_db):
    chellow.e.bill_importer.import_id = 0
    chellow.e.bill_importer.imports.clear()
    return create_app(testing=True)


@pytest.fixture
def raw_client(app):
    return app.test_client()


class CustomClient(FlaskClient):
    def open(self, *args, **kwargs):
        if "headers" in kwargs:
            headers = kwargs["headers"]
        else:
            headers = kwargs["headers"] = {}

        headers["Authorization"] = _basic_auth_str("admin@example.com", "admin")
        return super().open(*args, **kwargs)


@pytest.fixture
def sess(app):
    sess = Session()
    yield sess
    sess.close()


@pytest.fixture
def client(app, sess):
    app.test_client_class = CustomClient

    with app.test_client() as client:
        with app.app_context():
            user_role = UserRole.insert(sess, "editor")
            User.insert(sess, "admin@example.com", "admin", user_role, None)

            sess.execute(
                "INSERT INTO market_role (code, description) "
                "VALUES ('Z', 'Non-core Role')"
            )
            sess.execute(
                "INSERT INTO participant (code, name) " "VALUES ('NEUT', 'Neutral')"
            )
            sess.execute(
                "INSERT INTO party (market_role_id, participant_id, name, "
                "valid_from, valid_to, dno_code) "
                "VALUES (1, 1, 'Neutral Party', '2000-01-01', null, null)"
            )
            sess.execute(
                "INSERT INTO contract (name, charge_script, properties, "
                "state, market_role_id, party_id, start_rate_script_id, "
                "finish_rate_script_id) VALUES ('configuration', '{}', '{}', "
                "'{}', 1, 1, null, null)"
            )
            sess.execute(
                "INSERT INTO rate_script "
                "(contract_id, start_date, finish_date, "
                "script) VALUES (1, '2020-01-01', '2020-01-31', '{}')"
            )
            sess.execute(
                "UPDATE contract set start_rate_script_id = 1, "
                "finish_rate_script_id = 1 where id = 1;"
            )
            sess.commit()

        yield client


@pytest.fixture
def runner(app):
    return app.test_cli_runner()


@pytest.fixture
def jinja2_env():
    return Environment(
        loader=PackageLoader("../chellow/", "templates"),
        autoescape=select_autoescape(["html", "xml"]),
    )



import collections
import importlib
import threading
import traceback
from io import BytesIO
from pkgutil import iter_modules

from werkzeug.exceptions import BadRequest

import chellow
import chellow.e.bill_parsers.engie_xls
from chellow.models import (
    Batch,
    BatchFile,
    BillType,
    ReadType,
    Session,
    Supply,
    Tpr,
)
from chellow.utils import keydefaultdict, utc_datetime_now


import_id = 0
import_lock = threading.Lock()
imports = {}


def find_parser_names():
    return [name for _, name, _ in iter_modules(chellow.e.bill_parsers.__path__)]


class BillImport(threading.Thread):
    def __init__(self, batch):
        threading.Thread.__init__(self)
        global import_id
        self.import_id = import_id
        import_id += 1

        self.batch_id = batch.id
        self.successful_bills = []
        self.failed_bills = []
        self.log = collections.deque()
        self.bill_num = None
        self.parser = None

    def _log(self, msg):
        with import_lock:
            self.log.appendleft(
                f"{utc_datetime_now().strftime('%Y-%m-%d %H:%M:%S')} - {msg}"
            )

    def status(self):
        if self.is_alive():
            if self.bill_num is not None:
                return f"Inserting raw bills: I've reached bill number {self.bill_num}."
            elif self.parser is not None:
                return (
                    f"Parsing file: I've reached line number {self.parser.line_number}."
                )
            else:
                return "Running"
        else:
            return "Not running"

    def run(self):
        sess = None
        try:
            sess = Session()
            batch = Batch.get_by_id(sess, self.batch_id)

            bill_types = keydefaultdict(lambda k: BillType.get_by_code(sess, k))

            tprs = keydefaultdict(
                lambda k: None if k is None else Tpr.get_by_code(sess, k)
            )

            read_types = keydefaultdict(lambda k: ReadType.get_by_code(sess, k))

            for bf in (
                sess.query(BatchFile)
                .filter(BatchFile.batch == batch)
                .order_by(BatchFile.upload_timestamp)
            ):
                self.parser = _process_batch_file(sess, bf, self._log)
                for self.bill_num, raw_bill in enumerate(self.parser.make_raw_bills()):
                    if "error" in raw_bill:
                        self.failed_bills.append(raw_bill)
                    else:
                        try:
                            mpan_core = raw_bill["mpan_core"]
                            supply = Supply.get_by_mpan_core(sess, mpan_core)
                            with sess.begin_nested():
                                bill = batch.insert_bill(
                                    sess,
                                    raw_bill["account"],
                                    raw_bill["reference"],
                                    raw_bill["issue_date"],
                                    raw_bill["start_date"],
                                    raw_bill["finish_date"],
                                    raw_bill["kwh"],
                                    raw_bill["net"],
                                    raw_bill["vat"],
                                    raw_bill["gross"],
                                    bill_types[raw_bill["bill_type_code"]],
                                    raw_bill["breakdown"],
                                    supply,
                                )
                                for raw_read in raw_bill["reads"]:
                                    bill.insert_read(
                                        sess,
                                        tprs[raw_read["tpr_code"]],
                                        raw_read["coefficient"],
                                        raw_read["units"],
                                        raw_read["msn"],
                                        raw_read["mpan"],
                                        raw_read["prev_date"],
                                        raw_read["prev_value"],
                                        read_types[raw_read["prev_type_code"]],
                                        raw_read["pres_date"],
                                        raw_read["pres_value"],
                                        read_types[raw_read["pres_type_code"]],
                                    )
                                self.successful_bills.append(raw_bill)
                        except KeyError as e:
                            err = raw_bill.get("error", "")
                            raw_bill["error"] = err + " " + str(e)
                            self.failed_bills.append(raw_bill)
                        except BadRequest as e:
                            raw_bill["error"] = str(e.description)
                            self.failed_bills.append(raw_bill)

            if len(self.failed_bills) == 0:
                sess.commit()
                self._log(
                    "All the bills have been successfully loaded and attached "
                    "to the batch."
                )
            else:
                sess.rollback()
                self._log(
                    f"The import has finished, but there were {len(self.failed_bills)} "
                    f"failures, and so the whole import has been rolled back."
                )

        except BadRequest as e:
            sess.rollback()
            self._log(f"Problem: {e.description}")
        except BaseException:
            sess.rollback()
            self._log(f"I've encountered a problem: {traceback.format_exc()}")
        finally:
            if sess is not None:
                sess.close()

    def make_fields(self):
        with import_lock:
            fields = {
                "log": tuple(self.log),
                "is_alive": self.is_alive(),
                "importer_id": self.import_id,
            }
            if not self.is_alive():
                fields["successful_bills"] = self.successful_bills
                fields["failed_bills"] = self.failed_bills
            return fields


def _process_batch_file(sess, batch_file, log_f):
    data = batch_file.data
    parser_name = batch_file.parser_name

    if len(data) == 0:
        raise BadRequest("File has zero length")

    try:
        imp_mod = importlib.import_module(f"chellow.e.bill_parsers.{parser_name}")
    except ImportError:
        raise BadRequest(f"Can't find a parser with the name '{parser_name}'.")

    parser = imp_mod.Parser(BytesIO(data))
    log_f(f"Starting to parse the file with '{parser_name}'.")

    return parser


def start_bill_import(batch):
    with import_lock:
        bi = BillImport(batch)
        imports[bi.import_id] = bi
        bi.start()

    return bi.import_id


def get_bill_import_ids(batch):
    with import_lock:
        return [k for k, v in imports.items() if v.batch_id == batch.id]


def get_bill_import(id):
    with import_lock:
        return imports[id]



from datetime import datetime as Datetime

from dateutil.relativedelta import relativedelta

from flask import render_template


def do_get(sess):
    init = Datetime.utcnow()
    init = Datetime(init.year, init.month, 1) - relativedelta(months=1)
    return render_template("report_155.html", init=init)



from decimal import Decimal

import pytest

import sqlalchemy.exc

from werkzeug.exceptions import BadRequest

from chellow.models import (
    BillType,
    Comm,
    Contract,
    Cop,
    EnergisationStatus,
    Era,
    GspGroup,
    MarketRole,
    MeterPaymentType,
    MeterType,
    Mtc,
    MtcLlfc,
    MtcLlfcSsc,
    MtcLlfcSscPc,
    MtcParticipant,
    MtcSsc,
    Participant,
    Pc,
    Scenario,
    Site,
    Source,
    Ssc,
    Supply,
    Tpr,
    VoltageLevel,
    _jsonize,
    insert_bill_types,
    insert_comms,
    insert_cops,
    insert_energisation_statuses,
    insert_sources,
    insert_voltage_levels,
)
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_Bill_update(sess):
    """Make sure works if all instances in the Session have expired"""
    vf = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, vf, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess,
        "510",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        vf,
        None,
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "net")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        "773",
        dc_contract,
        "ghyy3",
        "e1msn",
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        {},
        "22 7867 6232 781",
        "510",
        supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    batch = supplier_contract.insert_batch(sess, "a", "ba")
    insert_bill_types(sess)
    bill_type = BillType.get_by_code(sess, "N")

    sess.commit()
    sess.expire_all()

    batch.insert_bill(
        sess,
        "xhgerol",
        "asdgh98g",
        to_utc(ct_datetime(2022, 3, 1)),
        to_utc(ct_datetime(2022, 1, 1)),
        to_utc(ct_datetime(2022, 1, 31, 23, 30)),
        Decimal("0"),
        Decimal("0.00"),
        Decimal("0.00"),
        Decimal("0.00"),
        bill_type,
        {},
        supply,
    )


def test_Era_update(sess):
    """Successful"""
    valid_from = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", valid_from, None, None)
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop", valid_from, None, None)
    participant.insert_party(sess, market_role_X, "Fusion", valid_from, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", valid_from, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, valid_from, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, valid_from, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", utc_datetime(2000, 1, 1), None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_contract = Contract.insert_supplier(
        sess, "Fusion Supplier 2000", participant, "", {}, valid_from, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", valid_from, None, "22")
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", utc_datetime(2000, 1, 1), None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", valid_from, None)
    mtc = Mtc.insert(sess, "845", False, True, valid_from, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        valid_from,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "510", "PC 5-8 & HH HV", voltage_level, False, True, valid_from, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, valid_from, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "net")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    imp_mpan_core = "22 7867 6232 781"
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        "773",
        dc_contract,
        "ghyy3",
        "e1msn",
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        {},
        imp_mpan_core,
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    era = supply.eras[0]
    msn = "e2msn"
    era.update(
        sess,
        era.start_date,
        None,
        mop_contract,
        "379540",
        dc_contract,
        "547yt",
        msn,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        {},
        imp_mpan_core,
        "510",
        imp_supplier_contract,
        "9745y6",
        361,
        None,
        None,
        None,
        None,
        None,
    )

    sess.commit()
    era = Era.get_by_id(sess, era.id)

    assert era.msn == msn


def test_Era_meter_category_dumb(sess):
    """Successful"""
    valid_from = utc_datetime(1996, 1, 1)
    site = Site.insert(sess, "CI017", "Water Works")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", valid_from, None, None)
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop", valid_from, None, None)
    participant.insert_party(sess, market_role_X, "Fusion", valid_from, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", valid_from, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    pc = Pc.insert(sess, "02", "nhh", utc_datetime(2000, 1, 1), None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_contract = Contract.insert_supplier(
        sess, "Fusion Supplier 2000", participant, "", {}, valid_from, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", valid_from, None, "22")
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", utc_datetime(2000, 1, 1), None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", valid_from, None)
    mtc = Mtc.insert(sess, "845", False, True, valid_from, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        valid_from,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess,
        "510",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        utc_datetime(1996, 1, 1),
        None,
    )
    insert_sources(sess)
    source = Source.get_by_code(sess, "net")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    ssc = Ssc.insert(sess, "0001", "All", True, utc_datetime(1996, 1), None)
    MtcLlfc.insert(sess, mtc_participant, llfc, valid_from, None)
    mtc_ssc = MtcSsc.insert(sess, mtc_participant, ssc, valid_from, None)
    mtc_llfc_ssc = MtcLlfcSsc.insert(sess, mtc_ssc, llfc, valid_from, None)
    MtcLlfcSscPc.insert(sess, mtc_llfc_ssc, pc, valid_from, None)
    imp_mpan_core = "22 7867 6232 781"
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        "773",
        dc_contract,
        "ghyy3",
        "e1msn",
        pc,
        "845",
        cop,
        comm,
        ssc,
        energisation_status,
        {},
        imp_mpan_core,
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    era = supply.eras[0]
    era.meter_category

    sess.commit()


def test_Mtc_find_by_code(sess):
    code = "034"
    valid_from = to_utc(ct_datetime(2000, 1, 1))
    Mtc.insert(
        sess,
        code,
        False,
        True,
        valid_from,
        None,
    )
    sess.commit()

    mtc = Mtc.find_by_code(sess, "34", valid_from)
    assert mtc.code == code


def test_MtcParticipant_find_by_values(sess):
    valid_from = to_utc(ct_datetime(2000, 1, 1))
    participant = Participant.insert(sess, "CALB", "AK Industries")
    code = "034"
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", utc_datetime(2000, 1, 1), None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", valid_from, None)
    mtc = Mtc.insert(
        sess,
        code,
        False,
        True,
        valid_from,
        None,
    )
    MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "an mtc",
        False,
        True,
        meter_type,
        meter_payment_type,
        1,
        utc_datetime(2000, 1, 1),
        None,
    )
    sess.commit()

    mtc_participant = MtcParticipant.find_by_values(sess, mtc, participant, valid_from)
    assert mtc_participant.mtc.code == code


def test_MtcSSC_get_by_values(sess):
    valid_from = to_utc(ct_datetime(2000, 1, 1))
    participant = Participant.insert(sess, "CALB", "AK Industries")
    code = "034"
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", valid_from, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", valid_from, None)
    mtc = Mtc.insert(
        sess,
        code,
        False,
        True,
        valid_from,
        None,
    )
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "an mtc",
        False,
        True,
        meter_type,
        meter_payment_type,
        1,
        utc_datetime(2000, 1, 1),
        None,
    )
    ssc = Ssc.insert(sess, "0001", "All", True, utc_datetime(1996, 1), None)
    sess.commit()

    with pytest.raises(
        BadRequest,
        match="For the participant CALB there isn't an MTC SSC with the MTC 034 and "
        "SSC 0001 at date 2000-01-01 00:00.",
    ):
        MtcSsc.get_by_values(sess, mtc_participant, ssc, valid_from)


def test_Era_init_llfc_valid_to(sess):
    """
    Error raised if LLFC finishes before the era
    """
    site = Site.insert(sess, "CI017", "Water Works")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(
        sess, market_role_Z, "None core", utc_datetime(2000, 1, 1), None, None
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(
        sess, market_role_M, "Fusion Mop Ltd", utc_datetime(2000, 1, 1), None, None
    )
    participant.insert_party(
        sess, market_role_X, "Fusion Ltc", utc_datetime(2000, 1, 1), None, None
    )
    participant.insert_party(
        sess, market_role_C, "Fusion DC", utc_datetime(2000, 1, 1), None, None
    )
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    pc = Pc.insert(sess, "00", "hh", utc_datetime(2000, 1, 1), None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    dno = participant.insert_party(
        sess, market_role_R, "WPD", utc_datetime(2000, 1, 1), None, "22"
    )
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", utc_datetime(2000, 1, 1), None)
    meter_payment_type = MeterPaymentType.insert(
        sess, "CR", "Credit", utc_datetime(1996, 1, 1), None
    )
    mtc = Mtc.insert(
        sess,
        "845",
        False,
        True,
        utc_datetime(1996, 1, 1),
        None,
    )
    MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        utc_datetime(1996, 1, 1),
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    dno.insert_llfc(
        sess,
        "510",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        to_utc(ct_datetime(2000, 5, 1)),
        to_utc(ct_datetime(2010, 5, 1)),
    )
    insert_sources(sess)
    source = Source.get_by_code(sess, "net")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    imp_mpan_core = "22 7867 6232 781"
    with pytest.raises(
        BadRequest,
        match="The imp line loss factor 510 is only valid until "
        "2010-05-01 00:00 but the era ends at 2011-01-01 00:00.",
    ):
        site.insert_e_supply(
            sess,
            source,
            None,
            "Bob",
            to_utc(ct_datetime(2002, 7, 1)),
            to_utc(ct_datetime(2011, 1, 1)),
            gsp_group,
            mop_contract,
            "773",
            dc_contract,
            "ghyy3",
            "e1msn",
            pc,
            "845",
            cop,
            comm,
            None,
            energisation_status,
            {},
            imp_mpan_core,
            "510",
            imp_supplier_contract,
            "7748",
            361,
            None,
            None,
            None,
            None,
            None,
        )


def test_Contract_get_next_batch_details(mocker):
    MockBatch = mocker.patch("chellow.models.Batch", autospec=True)
    MockBatch.contract = mocker.Mock()
    instance = mocker.Mock()

    batch_description = "A King"
    batch = mocker.Mock()
    batch.reference = "king-098"
    batch.description = batch_description

    returns = iter([batch])

    class Sess:
        def query(self, *args):
            return self

        def join(self, *args):
            return self

        def order_by(self, *args):
            return self

        def filter(self, *args):
            return self

        def first(self, *args):
            return next(returns)

    sess = Sess()
    ref, desc = Contract.get_next_batch_details(instance, sess)
    assert ref == "king-099"
    assert desc == batch_description


def test_Contract_get_next_batch_details__no_suffix(mocker):
    MockBatch = mocker.patch("chellow.models.Batch", autospec=True)
    MockBatch.contract = mocker.Mock()
    instance = mocker.Mock()

    batch_reference = "king"
    batch_description = "A King"
    batch = mocker.Mock()
    batch.reference = batch_reference
    batch.description = batch_description

    returns = iter([batch])

    class Sess:
        def query(self, *args):
            return self

        def join(self, *args):
            return self

        def order_by(self, *args):
            return self

        def filter(self, *args):
            return self

        def first(self, *args):
            return next(returns)

    sess = Sess()
    ref, desc = Contract.get_next_batch_details(instance, sess)
    assert ref == batch_reference
    assert desc == batch_description


def test_Tpr_get_by_code_not_found(sess):
    with pytest.raises(
        BadRequest,
        match="A TPR with code 'g' expanded to '0000g' can't be found.",
    ):
        Tpr.get_by_code(sess, "g")


def test_Tpr_get_by_code_found(sess):
    code = "00001"
    is_teleswitch = False
    is_gmt = True
    Tpr.insert(sess, code, is_teleswitch, is_gmt)
    sess.commit()

    tpr = Tpr.get_by_code(sess, code)
    assert tpr.code == code
    assert tpr.is_teleswitch == is_teleswitch
    assert tpr.is_gmt == is_gmt


def test_sql_insert_GExitZone(mocker, sess):
    with pytest.raises(
        sqlalchemy.exc.ProgrammingError,
        match='null value in column "g_ldz_id" violates not-null ' "constraint",
    ):
        sess.execute(
            "INSERT INTO g_exit_zone (id, code, g_ldz_id) VALUES "
            "(DEFAULT, 'E1', null)"
        )


def test_Supply_get_by_MPAN_core(sess):
    mpan_core = "22 1737 1873 221"

    with pytest.raises(
        BadRequest, match=f"The MPAN core {mpan_core} is not set up in Chellow."
    ):
        Supply.get_by_mpan_core(sess, mpan_core)


def test_Supply_insert_era_at(sess):
    """Where an era is inserted in the last HH of another era, check
    the template era is the one at the insertion date.
    """
    valid_from = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", valid_from, None, None)
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop", valid_from, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", valid_from, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", valid_from, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, valid_from, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    pc = Pc.insert(sess, "00", "hh", utc_datetime(2000, 1, 1), None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", valid_from, None, "22")
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", utc_datetime(2000, 1, 1), None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", valid_from, None)
    mtc = Mtc.insert(sess, "845", False, True, valid_from, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        utc_datetime(1996, 1, 1),
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess,
        "510",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        utc_datetime(1996, 1, 1),
        None,
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, valid_from, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "net")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    era1_msn = "e1msn"
    imp_mpan_core = "22 7867 6232 781"
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        "773",
        dc_contract,
        "ghyy3",
        era1_msn,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        {},
        imp_mpan_core,
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    era1 = supply.eras[0]
    era2_start_date = utc_datetime(2009, 7, 31, 23, 30)
    era2 = supply.insert_era_at(sess, era2_start_date)
    era2_msn = "e2msn"
    era2.update(
        sess,
        era2_start_date,
        None,
        mop_contract,
        "379540",
        dc_contract,
        "547yt",
        era2_msn,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        {},
        imp_mpan_core,
        "510",
        imp_supplier_contract,
        "9745y6",
        361,
        None,
        None,
        None,
        None,
        None,
    )

    sess.commit()

    start_date = utc_datetime(2009, 7, 31, 23, 00)
    era3 = supply.insert_era_at(sess, start_date)
    assert era3.msn == era1.msn


def test_Scenario_init(sess):
    name = "scenario_bau"
    properties = {
        "local_rates": {},
        "scenario_start_year": 2015,
        "scenario_start_month": 6,
        "scenario_duration": 1,
    }

    with pytest.raises(BadRequest, match="The 'local_rates' must be a list."):
        Scenario(name, properties)


def test_jsonize():
    val = {1, None}
    actual = _jsonize(val)
    assert actual == [1, None]
